import numpy as np

from abc import abstractmethod, ABCMeta
from dataclasses import dataclass
from enum import Enum
from typing import Optional, TypeVar, Generic, List

from dataspree.inference.core.item import Item


class InferenceMode(str, Enum):
    """
    Enumeration for different inference modes.

    Attributes:
        WHOLE_IMAGE (str): Inference mode for processing the whole image.
        ALL_ROIS (str): Inference mode for processing all regions of interest (ROIs).
        SELECTED_ROIS (str): Inference mode for processing selected ROIs.
    """
    WHOLE_IMAGE = 'whole_image'
    ALL_ROIS = 'all_rois'
    SELECTED_ROIS = 'selected_rois'


@dataclass
class PreprocessedContent:
    """
    Content used for performing inference.

    Attributes:
        inference_batches (np.ndarray): Image data for inference.
            Shape: (number of rois, number of crops, number of tta, network input specific).

        image_context (Optional[np.ndarray]): Image context used for visualizing heatmaps.

    Properties:
        n_roi (int): Number of regions of interest (ROIs) in the preprocessed content.
        n_crop (int): Number of crops in the preprocessed content.
        n_tta (int): Number of Test Time Augmentations (TTA) applied to the preprocessed content.
    """

    inference_batches: np.ndarray
    image_context: Optional[np.ndarray]

    @property
    def n_roi(self) -> int:
        """Number of regions of interest (ROIs) in the preprocessed content."""
        return self.inference_batches.shape[0]

    @property
    def n_crop(self) -> int:
        """Number of crops in the preprocessed content."""
        return self.inference_batches.shape[1]

    @property
    def n_tta(self) -> int:
        """Number of Test Time Augmentations (TTA) applied to the preprocessed content."""
        return self.inference_batches.shape[2]


T = TypeVar('T', bound=PreprocessedContent)


class BaseModel(Generic[T], metaclass=ABCMeta):
    """
    Abstract base class defining the interface for models used in the inference pipeline.

    This class provides a framework for implementing models that are compatible with the inference pipeline.
    Subclasses must implement specific methods to perform preprocessing, prediction, and postprocessing.
    """

    @abstractmethod
    def preprocess(self, processed_input: Item) -> T:
        """
        Processes the item and extracts `PreprocessedContent` on which inference can be performed.

        This step is performed first in the inference pipeline.

        Args:
            processed_input (Item): The input item to be processed.

        Returns:
            T: The preprocessed content to be used for inference.

        Raises:
            InferenceError: If the input item cannot be processed for inference, the configuration is wrong or no data
                            exists.
        """

        pass

    @abstractmethod
    def tta_apply(self, processed_input: T) -> T:
        """
        Apply test time augmentations (TTA) to the preprocessed input. This function copies the preprocessed
        input and applies different augmentations, which can later be reversed and merged into a single output map.

        This step is performed on the output of `preprocess`.

        Args:
            processed_input (numpy.ndarray): The input data after preprocessing.
                Shape: (number of rois, number of crops, 1, ...)

        Returns:
            numpy.ndarray: The augmented input data.
                Shape: (number of rois, number of crops, number of tta, ...)
        """

    @abstractmethod
    def predict(self, model_input: np.ndarray) -> List[np.ndarray]:
        """
        Generates predictions for the input batch using the model.

        This step is performed on the output of `tta_apply`.

        Args:
            model_input (np.ndarray): The input data after applying Test Time Augmentation (TTA).
                Shape: (number of rois, number of crops, number of tta, ...)

        Returns:
            List[np.ndarray]: The predictions from the model.
                Shape: (number of rois, number of crops, number of tta, ...), reshaped to associate each output with
                the corresponding region of interest (ROI), crop, and test time augmentation (TTA).
        """

    @abstractmethod
    def compute_class_activation_map(self, model_output: List[np.ndarray]) -> np.ndarray:
        """
        Compute the class activation map from model output.

        This step is performed on the output of `predict`, `tta_rectify` or `tta_merge`.

        Args:
            model_output (List[np.ndarray]): The output of the mode.
                Shape: (number of rois, number of crops, number of tta, ...).

        Returns:
            np.ndarray: A NumPy array representing the computed class activation map.

        Raises:
            InferenceError: If the class activation map cannot be generated.

        """
        ...

    @abstractmethod
    def tta_rectify(self, processed_input: T, model_output: List[np.ndarray]) -> List[np.ndarray]:
        """
        Applies a transformation to the predicted output maps that reverses the Test Time Augmentations (TTA).

        This step is performed on the output of `predict`.

        Args:
            processed_input (T): The input data after applying TTA.
                Shape: (number of rois, number of crops, 1, ...)

            model_output (List[np.ndarray]): The output of the model, with the shape adjusted to match the
                                             processed input.
                Shape: (number of rois, number of crops, number of tta, ...)

        Returns:
            List[np.ndarray]: The rectified predictions, corresponding to the original input dimensions.
                Shape: (number of rois, number of crops, number of tta, ...)
        """

    @abstractmethod
    def tta_merge(self, model_output: List[np.ndarray]) -> List[np.ndarray]:
        """
        Merges the rectified output maps that correspond to different Test Time Augmentations (TTA) into a single map.

        This step is performed on the output of `tta_rectify`.

        Args:
            model_output (List[np.ndarray]): The rectified and shape-adjusted output of the model.
                Shape: (number of rois, number of crops, number of tta, ...)

        Returns:
            List[np.ndarray]: The merged outputs.
                Shape: (number of rois, number of crops, 1, ...)
        """

    @abstractmethod
    def postprocess(self, model_output: List[np.ndarray], processed_input: T, item: Item) -> None:
        """
        Postprocess the output of the model and populate the content into the given item.
        This method finalizes the model's predictions by applying any necessary transformations and then
        integrating the results back into the provided item.

        This step is performed on the output of `tta_merge`.

        Args:
            model_output (List[np.ndarray]): The shape-adjusted output of the model with merged TTA.
                Shape: (number of rois, number of crops, 1, ...)

            processed_input (T): The input data that was originally processed and augmented.

            item (Item): The item to be populated with the final predictions.
        """

    @abstractmethod
    def destroy(self) -> None:
        """
        Release the allocated resources.
        """
