from abc import ABC, abstractmethod
from typing import final

from dataspree.inference.core.signals import BaseSignal

import logging

logger = logging.getLogger(__name__)


class SignalHandler(ABC):
    """ An instance of a signal handler subclass works as a node in the signal routing graph.
    It can receive signals and propagate them to others.
    The exact behavior is defined in the implementation of the class.
    """

    def __init__(self):
        ABC.__init__(self)
        # Identifier used in order to check equality of objects derived from the instance that was deep or shallow
        # copied along the way
        self.__og_id: int = id(self)

    @final
    def handle_signal(self, signal: BaseSignal) -> None:
        if self.is_active():
            logger.debug(f'{self} received {signal=}.')
            if not signal.loop_finished(self):
                self.process_signal(signal)
                signal.transmit(self)

    def propagate_signal(self, signal: BaseSignal) -> None:
        """ Propagates the signal to the connected transmission nodes.
        When creating a new signal, this method can be applied to propagate the signals as defined in its implementation.

        Example by initially emitting a TerminatedNodeSignal in a subclass of SignalHandler:

            >>> self.propagate_signal(TerminatedNodeSignal(self))

        The implementation of this method defines how and to which next signal handlers the signal is propagated
        from the current instance. By default, if not defined otherwise in a subclass, the signal is not propagated.

        :param signal:  the signal instance that is supposed to be propagated.
        """
        pass

    def process_signal(self, signal: BaseSignal) -> None:
        """ Defines the behavior of the signal handler when receiving a certain signal.

        To handle specific signals which by definition are differentiated by their type and a subclass of BaseSignal,
        the process_signal method can be implemented by the subclass of SignalHandler with the `@singledispatchmethod`
        decorator.
        This way, python dispatching (https://docs.python.org/3/library/functools.html) can be used to handle
        different signal types with in separate methods with the `@process_signal.register` decorator.
        A different name than `process_signal` should be used for the registered methods (e.g. _ is typically used).

        Example:
        --------------------

            >>> @singledispatchmethod
            >>> def process_signal(self, signal: BaseSignal) -> None:
            >>>     print(f'{self} received {signal}')
            >>>
            >>> @process_signal.register
            >>> def _(self, signal: TerminatedNodeSignal) -> None:
            >>>     print(f'{self} received {signal}')

        Important:
        --------------------

        In external plugins, custom signal classes have to be defined in a dedicated module and imported in the plugin
        module in order to be able to use the `@process_signal.register` decorator appropriately.

        Note:
        --------------------

        @dev implementing `@singledispatchmethod.process_signal` here in the SignalHandler, and decorating
        the implementations with `@SignalHandler.process_signal.register` in the subclasses is not possible.
        This causes the signal dispatcher to be not tied to the instance.
        If there are multiple classes that register a dispatch method, this would cause all of them to be active
        whenever one of them is supposed to be invoked.

        :param signal:  the signal that is supposed to be processed and/or dispatched.
        """
        pass

    def __eq__(self, other) -> bool:
        return isinstance(other, SignalHandler) and self.__og_id == other.__og_id

    @abstractmethod
    def is_active(self) -> bool:
        pass
