import time

import numpy as np

from typing import Any, Dict, List, Tuple, Optional

from albumentations import DualTransform
import cv2


class ChannelAssembler(DualTransform):
    """
    The ChannelAssembler takes a list of channels and a structured array to compose the unstructured image array.

    The list of channels must be as follows:
    >>> [
    >>>    {'channel': {'channel_type': {'selection': 'red'}}},
    >>>    {'channel': {'channel_type': {'selection': 'green'}}},
    >>>    {'channel': {'channel_type': {'selection': 'blue'}}}
    >>> ]
    """

    def __init__(self, channels: List[Dict[str, Any]], always_apply: bool, p: float) -> None:
        super().__init__(always_apply=always_apply, p=p)
        self.channels = channels
        self.__last_img = None
        self.__last_result = None

    @staticmethod
    def retrieve_channel(channel_name: str, img):
        image_channel: np.ndarray
        available_channel_names: Tuple[str, ...] = img.dtype.names
        if channel_name in available_channel_names:
            return img[channel_name]

        elif channel_name == 'gray' or channel_name == 'a' or channel_name == 'grey':
            return ((img['red'].astype(np.float32) + img['green'].astype(np.float32)
                     + img['blue'].astype(np.float32)) / 3).astype(img['red'].dtype)

        raise ValueError(f'Channel {channel_name} not available. Available channels: '
                         f'{", ".join(available_channel_names)}.')

    def apply(self, img: np.ndarray, **_params) -> np.ndarray:
        if img is self.__last_img:
            return self.__last_result

        is_structured_array: bool = img.dtype.names is not None
        if not is_structured_array:
            return img

        # target in channel first layout

        init_channel_first = False
        if init_channel_first:
            output_array = np.empty((len(self.channels), *img.shape), dtype=np.float32)
        else:
            output_array = np.empty((*img.shape, len(self.channels)), dtype=np.float32)

        # Scale to 0-1 and assign to output array
        for channel_id, channel in enumerate(self.channels):
            if init_channel_first:
                channel_output = output_array[channel_id]
            else:
                channel_output = output_array[..., channel_id]
            channel_configuration = channel['channel']
            channel_name: str = channel_configuration['channel_type']['selection']
            if channel_name == 'custom':
                channel_name = channel_configuration['channel_type']['name']

            # Collect channel from image or compose it from a set of available options.
            image_channel = self.retrieve_channel(channel_name, img)

            is_integer_dtype: bool = np.issubdtype(image_channel.dtype, np.integer)
            is_floating_point: bool = not is_integer_dtype
            # is_signed_integer_dtype: bool = np.issubdtype(image_channel.dtype, np.signedinteger)
            # is_unsigned_integer: bool = is_integer_dtype and not is_signed_integer_dtype

            channel_operations = channel_configuration.get('enable_operations', dict()).get('operations', dict())

            # Defaults if not configured
            enabled_operation = channel_operations.get('selection')
            if not enabled_operation:
                channel_operations['min'] = 0.0
                channel_operations['max'] = 1.0
                enabled_operation = 'relative_cap' if is_integer_dtype else 'absolute_cap'

            min_value: float
            max_value: float
            need_clipping: bool

            if enabled_operation == 'absolute_cap':
                min_value = float(channel_operations['min'])
                max_value = float(channel_operations['max'])
                need_clipping = True

            elif enabled_operation == 'relative_cap':
                if is_floating_point:
                    raise ValueError(
                        '"Relative Cap" can only be used on integer channels, not on channels of '
                        f'dtype {image_channel.dtype}. '
                        f'Processing channel {channel_name} failed for item.')

                min_per_datatype = np.iinfo(image_channel.dtype).min
                max_per_datatype = np.iinfo(image_channel.dtype).max
                range_datatype = max_per_datatype - min_per_datatype

                rel_min: float = float(channel_operations.get('min', 0.0))
                rel_max: float = float(channel_operations.get('max', 1.0))
                if rel_min == 0.0 and rel_max == 1.0:
                    need_clipping = False
                    min_value = float(min_per_datatype)
                    max_value = float(max_per_datatype)
                else:
                    need_clipping = True
                    min_value = rel_min * range_datatype + min_per_datatype
                    max_value = rel_max * range_datatype + min_per_datatype

            elif enabled_operation == 'normalize':
                min_value = float(np.min(image_channel))
                max_value = float(np.max(image_channel))
                need_clipping = False

            else:
                raise ValueError(f'Unknown operation "{enabled_operation}" for channel {channel_name}.')

            # CV2 is significantly (1-5) x quicker on uint types with just normalize, but slower (1-5) x on float32.
            if enabled_operation == 'normalize' and not is_floating_point:
                cv2.normalize(image_channel, channel_output,
                              alpha=-1.0, beta=1.0, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F)

            else:
                if max_value <= min_value:
                    channel_output.fill(0.0)

                else:
                    if min_value != 0:
                        np.subtract(image_channel, min_value, out=channel_output)
                        np.multiply(channel_output, 2.0 / (max_value - min_value), out=channel_output)
                    else:
                        np.multiply(image_channel, 2.0 / (max_value - min_value), out=channel_output)

                    np.subtract(channel_output, 1, out=channel_output)
                    if need_clipping:
                        np.clip(channel_output, -1.0, 1.0, out=channel_output)

        if init_channel_first:
            return np.moveaxis(output_array, 0, -1)

        self.__last_result, self.__last_img = output_array.copy(), img
        return output_array

    def apply_to_bbox(self, bbox, **_params):
        return bbox

    def apply_to_keypoint(self, keypoint, **_params):
        return keypoint

    def get_transform_init_args_names(self) -> Tuple[str]:
        return ('channels',)

    @classmethod
    def get_class_fullname(cls):
        return cls.__name__
