from typing import Any, Dict
from albumentations import BaseCompose

import random
import numpy as np


class OneOfRetainProb(BaseCompose):
    """Select one of transforms to apply. Selected transform will be called with `force_apply=True`.
    Transforms probabilities will be normalized to one 1, so in this case transforms probabilities works as weights.
    If the sum of all transforms is less than 1, the probability to select one transform is adjusted to maintain the
    original probabilities.
    Example:
        Transform A.p = 0.1
        Transform B.p = 0.2
        The probability of OneOf([A, B], p=1.0) is adjusted from 1.0 to 0.3 (A.p + B.p), to maintain the original
        probabilities of selecting transform A (0.1) and B (0.2).
        The probability of OneOf([A, B], p=0.5) is adjusted from 0.5 to 0.15 accordingly.

    Args:
        transforms (list): list of transf
        g selected transform. Default: 0.5.
    """

    def __init__(self, transforms, p: float = 0.5):
        super(OneOfRetainProb, self).__init__(transforms, p)
        transforms_ps = [t.p for t in self.transforms]
        s = sum(transforms_ps)
        if s < 1.0:
            self.p *= s
        self.transforms_ps = [t / s for t in transforms_ps]

    def __call__(self, *args, force_apply: bool = False, **data) -> Dict[str, Any]:
        if self.replay_mode:
            for t in self.transforms:
                data = t(**data)
            return data

        if self.transforms_ps and (force_apply or random.random() < self.p):
            random_state = np.random.RandomState(random.randint(0, 2 ** 32 - 1))
            t = random_state.choice(self.transforms.transforms, p=self.transforms_ps)
            data = t(force_apply=True, **data)
        return data

    @classmethod
    def get_class_fullname(cls):
        return cls.__name__
