from albumentations import PadIfNeeded
from typing import Union, Optional

import cv2


class RelativePadding(PadIfNeeded):
    def __init__(self, aspect_ratio: float,
                 position: Union[PadIfNeeded.PositionType, str] = PadIfNeeded.PositionType.CENTER,
                 border_mode=cv2.BORDER_CONSTANT, value=0, mask_value=None, always_apply=False, p=1.0):
        """
        :param aspect_ratio:   width / height
        """
        super(RelativePadding, self).__init__(min_height=1, min_width=1, position=position, border_mode=border_mode,
                                              value=value, mask_value=mask_value, always_apply=always_apply, p=p)
        if aspect_ratio <= 0:
            raise ValueError('Aspect ratio has to be > 0')

        self.aspect_ratio: float = aspect_ratio
        self.__last_input_height: int = 0
        self.__last_input_width: int = 0

    def update_params(self, params, **kwargs):
        params = super(RelativePadding, self).update_params(params, **kwargs)
        self.__last_input_height = params['rows']
        self.__last_input_width = params['cols']

        if self.__last_input_height >= self.__last_input_width:
            self.min_height = self.__last_input_height
            self.min_width = max(int(round(self.min_height * self.aspect_ratio)), 1)
        else:
            self.min_width = self.__last_input_width
            self.min_height = max(int(round(self.min_width / self.aspect_ratio)), 1)

        return super(RelativePadding, self).update_params(params, **kwargs)

    def __check_after_update(self, rows: Optional[int] = None, cols: Optional[int] = None, **_kwargs):
        if rows is None and cols is None:
            raise NotImplementedError('Rows and Columns have to be provided by a compatible version of Albumentations. '
                                 'Make sure you use the recommended version of Albumentations and report this '
                                 'incompatibility to a DataSpree developer.')
        else:
            if rows != self.__last_input_height:
                raise NotImplementedError(f'Rows {rows} does not match expected height {self.__last_input_height}. '
                                     f'The current version of Albumentations is incompatible with the '
                                     f'{type(self).__name__} augmentation.'
                                     'Make sure that you use the recommended version of Albumentations and report '
                                     'this incompatibility to a DataSpree developer.')
            if cols != self.__last_input_width:
                raise NotImplementedError(f'Rows {cols} does not match expected height {self.__last_input_width}'
                                     f'The current version of Albumentations is incompatible with the '
                                     f'{type(self).__name__} augmentation.'
                                     'Make sure that you use the recommended version of Albumentations and report '
                                     'this incompatibility to a DataSpree developer.')

    def apply(self, img, **params):
        self.__check_after_update(**params)
        res = super(RelativePadding, self).apply(img, **params)
        return res

    def apply_to_bbox(self, bbox, **params):
        self.__check_after_update(**params)
        return super(RelativePadding, self).apply_to_bbox(bbox, **params)

    def apply_to_keypoint(self, keypoint, **params):
        self.__check_after_update(**params)
        return super(RelativePadding, self).apply_to_keypoint(keypoint, **params)

    def get_transform_init_args_names(self):
        return 'aspect_ratio', 'border_mode', 'value', 'mask_value'

    @classmethod
    def get_class_fullname(cls):
        return cls.__name__
