"""Components module for handling pixel format components in image processing.

This module defines classes, methods, and utilities for managing pixel components,
channels, and their associated properties in various formats and standards.
"""

import numpy as np
from abc import ABCMeta, abstractmethod
from collections.abc import Sequence, Iterator
from enum import Enum
from dataclasses import dataclass
from typing import Union, Optional, get_args, Any, TypeVar, cast, TypeGuard
from logging import getLogger

from dataspree.inference.sensor.data.pixel_format.utils import join_designators, all_implementations
from dataspree.inference.sensor.data.pixel_format.number_of_bits import Size

logger = getLogger(__name__)


class CompoundChannel(Enum):
    """Enumeration of compound channels for pixel formats."""

    RAW = 'Raw'
    MONO = 'Mono'
    CONFIDENCE = 'Confidence'
    RED = 'R'
    GREEN = 'G'
    BLUE = 'B'
    LUMA = 'Y'
    ALPHA = 'a'
    ALPHA_2 = 'A'
    HUE = 'H'
    SATURATION = 'S'
    VALUE = 'V'  # or color difference for YUV
    CHROMA = 'U'
    WHITE = 'W'
    INTENSITY = 'I'
    BLUE_DIFFERENCE = 'Cb'
    RED_DIFFERENCE = 'Cr'
    CYAN = 'C'
    MAGENTA = 'M'
    YELLOW = 'Ye'
    INFRARED = 'Ir'


single_channel_priority_mapping: dict[CompoundChannel, float] = {
    CompoundChannel.INTENSITY: 1,
    CompoundChannel.LUMA: 1,
    CompoundChannel.MONO: 1,
    CompoundChannel.VALUE: 2,
    CompoundChannel.GREEN: 2,
    CompoundChannel.RED: 3,
    CompoundChannel.BLUE: 3,
    CompoundChannel.SATURATION: 2,
}


class CoordinateChannel(Enum):
    """Enumeration of coordinate channels for 3D spatial data."""

    COORD3D_A = 'A'
    COORD3D_B = 'B'
    COORD3D_C = 'C'


_custom_channels: list[str] = []

Channel = Union[CompoundChannel, CoordinateChannel, str]

_reverse_channel_mapping: dict[str, Channel] = {f.value: f for f in CompoundChannel}

if len(_reverse_channel_mapping) != len(CompoundChannel):
    raise ValueError('Invalid channel mapping.')


def is_channel(potential_channel: Any) -> TypeGuard[Channel]:
    """Checks if a given object is a valid channel.

    Args:
        potential_channel: The object to validate.

    Returns:
        bool: True if the object is a valid channel, False otherwise.
    """
    return any(isinstance(potential_channel, arg) for arg in get_args(Channel))


def channel_name(channel: Channel) -> str:
    """Retrieves the string representation of a channel.

    Args:
        channel (Channel): The channel to convert.

    Returns:
        str: The string representation of the channel.
    """
    if isinstance(channel, str):
        return channel
    if isinstance(channel, CompoundChannel):
        return channel.value
    return channel.value


@dataclass
class ChannelProperty:
    """Represents properties of a channel.

    Attributes:
        combinable (bool): Whether the channel can be combined with others.
    """

    combinable: bool


channel_properties: dict[Channel, ChannelProperty] = {
    CompoundChannel.RAW: ChannelProperty(False),
    CompoundChannel.MONO: ChannelProperty(False),
    CompoundChannel.CONFIDENCE: ChannelProperty(False),
    CoordinateChannel.COORD3D_A: ChannelProperty(True),
    CoordinateChannel.COORD3D_B: ChannelProperty(True),
    CoordinateChannel.COORD3D_C: ChannelProperty(True),
    CompoundChannel.RED: ChannelProperty(True),
    CompoundChannel.GREEN: ChannelProperty(True),
    CompoundChannel.BLUE: ChannelProperty(True),
    CompoundChannel.LUMA: ChannelProperty(True),
    CompoundChannel.ALPHA: ChannelProperty(True),
    CompoundChannel.ALPHA_2: ChannelProperty(True),
    CompoundChannel.HUE: ChannelProperty(True),
    CompoundChannel.SATURATION: ChannelProperty(True),
    CompoundChannel.VALUE: ChannelProperty(True),
    CompoundChannel.CHROMA: ChannelProperty(True),
    CompoundChannel.WHITE: ChannelProperty(True),
    CompoundChannel.INTENSITY: ChannelProperty(True),
    CompoundChannel.BLUE_DIFFERENCE: ChannelProperty(True),
    CompoundChannel.RED_DIFFERENCE: ChannelProperty(True),
    CompoundChannel.CYAN: ChannelProperty(True),
    CompoundChannel.MAGENTA: ChannelProperty(True),
    CompoundChannel.YELLOW: ChannelProperty(True),
    CompoundChannel.INFRARED: ChannelProperty(True),
}

if len(channel_properties) != len(CompoundChannel) + len(CoordinateChannel):
    raise ValueError


def register_custom_channel(name: str, channel_property: ChannelProperty):
    """Registers a custom channel with its properties.

    Args:
        name (str): The name of the custom channel.
        channel_property (ChannelProperty): The properties of the channel.

    Raises:
        ValueError: If the channel name is invalid or already registered.
    """
    if name in _custom_channels:
        raise ValueError

    if name in channel_properties:
        raise ValueError

    if name in _reverse_channel_mapping:
        raise ValueError

    _custom_channels.append(name)
    channel_properties[name] = channel_property
    _reverse_channel_mapping[name] = name


def get_channels() -> Sequence[Channel]:
    """Gets a list of all registered channels.

    Returns:
        Sequence[Channel]: A list of all channels, including custom and predefined.
    """
    custom_channels: list[Channel] = list(_custom_channels)

    return custom_channels + list(CompoundChannel)


T = TypeVar('T', bound='Channels')


class Channels(metaclass=ABCMeta):
    """Abstract base class for handling pixel format channels."""

    @abstractmethod
    def name(self, indices: Optional[list[int]] = None) -> str:
        """Gets the name of the channel sequence.

        Args:
            indices (Optional[list[int]]): Indices to filter specific channels.

        Returns:
            str: The concatenated channel name.
        """

    @abstractmethod
    def __len__(self) -> int:
        """Gets the number of channels.

        Returns:
            int: The number of channels.
        """

    @abstractmethod
    def __getitem__(self, item: int) -> Channel:
        """Gets the channel at the specified index.

        Args:
            item (int): The index of the desired channel.

        Returns:
            Channel: The channel at the specified index.
        """

    def __iter__(self) -> Iterator[Channel]:
        """Iterator implementation."""
        for i in range(len(self)):
            yield self[i]

    @staticmethod
    def create(
        channel_start: str, allow_repeat: bool = False
    ) -> list[tuple['Channels', Optional['ColorSpaceStandard'], str]]:
        """Creates channel instances from a starting designation string.

        Args:
            channel_start (str): The starting designation string for channels.
            allow_repeat (bool): Whether duplicate channels are allowed.

        Returns:
            list[tuple['Channels', Optional['ColorSpaceStandard'], str]]: Parsed channels, optional color space
                                                                          standards, and remaining substrings.
        """
        possible_channels: list[tuple[Channels, str]] = []
        for implementation in all_implementations(Channels):
            try:
                possible_channels.extend(implementation.from_designation(channel_start, allow_repeat))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {implementation.__name__}')

        compos: list[tuple[Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, substring in possible_channels:
            optional_standards = ColorSpaceStandard.create(substring, channels)
            for cs, remaining_substring in optional_standards:
                compos.append((channels, cs, remaining_substring))
        return compos

    @classmethod
    @abstractmethod
    def from_designation(cls: type[T], channel_start: str, allow_repeat: bool = False) -> list[tuple[T, str]]:
        """Parses a designation string into channels.

        Args:
            channel_start (str): The start of the channel designation.
            allow_repeat (bool): Whether repeated channels are allowed.

        Returns:
            list[tuple['Channels', str]]: A list of channel instances and remaining substrings.
        """


CCT = TypeVar('CCT', bound='CompoundChannels')


class CompoundChannels(Channels):
    """Represents a compound set of pixel format channels."""

    def __init__(self, channels: Sequence[Channel], allow_repeat: bool = False) -> None:
        """Initializes a compound channel set.

        Args:
            channels (Sequence[Channel]): The list of channels.
            allow_repeat (bool): Whether duplicate channels are allowed.

        Raises:
            ValueError: If invalid channels or duplicates are encountered.
        """
        self.channels: Sequence[Channel] = channels

        for channel in self.channels:
            if channel not in channel_properties:
                raise ValueError

            channel_property = channel_properties[channel]
            if not channel_property.combinable and len(self.channels) > 1:
                raise ValueError

        if len(set(self.channels)) != len(self.channels) and not allow_repeat:
            raise ValueError('Repeat channel.')

    def __len__(self) -> int:
        """Gets the number of channels in the compound channel set.

        Returns:
            int: The number of channels.
        """
        return len(self.channels)

    def __getitem__(self, item: int) -> Channel:
        """Gets the channel at the specified index.

        Args:
            item (int): The index of the desired channel.

        Returns:
            Channel: The channel at the specified index.
        """
        return self.channels[item]

    def name(self, indices: Optional[list[int]] = None) -> str:
        """Gets the concatenated name of the channel sequence.

        Args:
            indices (Optional[list[int]]): A list of indices to filter specific channels.

        Returns:
            str: The concatenated name of the channels.
        """
        indices = list(range(len(self.channels))) if indices is None else indices
        return ''.join(map(channel_name, [self.channels[i] for i in indices]))

    @classmethod
    def from_designation(cls: type[CCT], channel_start: str, allow_repeat: bool = False) -> list[tuple[CCT, str]]:
        """Parses a designation string into compound channels.

        Args:
            channel_start (str): The start of the channel designation.
            allow_repeat (bool): Whether repeated channels are allowed.

        Returns:
            list[tuple[CompoundChannels, str]]: A list of channel instances and remaining substrings.
        """
        if not channel_start:
            return []

        if channel_start[0] == '_':
            channel_start = channel_start[1:]

        channels: list[tuple[CCT, str]] = []

        for channel_identifier, channel in _reverse_channel_mapping.items():
            if channel_start.startswith(channel_identifier):
                new_start = channel_start[len(channel_identifier) :]
                channels.append((cast(CCT, CompoundChannels([channel], allow_repeat=allow_repeat)), new_start))
                for lc, start in cls.from_designation(new_start, allow_repeat):
                    channels.append(
                        (cast(CCT, CompoundChannels([channel, *lc.channels], allow_repeat=allow_repeat)), start)
                    )

        return channels

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two CompoundChannel objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, CompoundChannels) and self.channels == other.channels


class Coord3d(CompoundChannels):
    """Represents 3D spatial coordinate channels."""

    def __init__(self, channels: Sequence[Channel], allow_repeat: bool = False) -> None:
        """Initializes a compound channel set.

        Args:
            channels (Sequence[Channel]): The list of channels to include.
            allow_repeat (bool): Whether duplicate channels are allowed.

        Raises:
            ValueError: If invalid channels or duplicates are encountered.
        """
        CompoundChannels.__init__(self, channels, allow_repeat=allow_repeat)
        for channel in channels:
            if channel not in (CoordinateChannel.COORD3D_A, CoordinateChannel.COORD3D_B, CoordinateChannel.COORD3D_C):
                raise ValueError

    def name(self, indices: Optional[list[int]] = None):
        """Generates the name of the Coord3D channel sequence.

        Args:
            indices (Optional[list[int]]): Indices to filter specific channels. If None, all channels are used.

        Returns:
            str: The formatted name of the Coord3D channel sequence.
        """
        return f'Coord3D_{CompoundChannels.name(self, indices)}'

    @classmethod
    def from_designation(
        cls: type['Coord3d'], channel_start, allow_repeat: bool = False
    ) -> list[tuple['Coord3d', str]]:
        """Parses a designation string into coord3d compound channels.

        Args:
            channel_start (str): The start of the channel designation.
            allow_repeat (bool): Whether repeated channels are allowed.

        Returns:
            list[tuple[Coord3d, str]]: A list of channel instances and remaining substrings.
        """
        if not channel_start:
            return []

        if channel_start[0] == '_':
            channel_start = channel_start[1:]

        if not channel_start.startswith('Coord3D_'):
            return []

        channel_start = channel_start[8:]
        channels: list[CoordinateChannel] = []
        for c in channel_start:
            if c == 'A':
                channels.append(CoordinateChannel.COORD3D_A)
            elif c == 'B':
                channels.append(CoordinateChannel.COORD3D_B)
            elif c == 'C':
                channels.append(CoordinateChannel.COORD3D_C)
            else:
                break

        if not len(channels):
            return []

        return [(Coord3d(channels, allow_repeat=allow_repeat), channel_start[len(channels) :])]


U = TypeVar('U', bound='ColorSpaceStandard')


class ColorSpaceStandard(metaclass=ABCMeta):
    """Abstract base class for color space standards."""

    def __init__(self, name: str) -> None:
        """Initializes the ITU-R BT.601 color space standard."""
        self.name: str = name

    @classmethod
    @abstractmethod
    def initialize(cls: type[U], channels: Channels) -> U:
        """Initializes a color space standard for the given channels.

        Args:
            channels (Channels): The channel set for the standard.

        Returns:
            ColorSpaceStandard: An initialized color space standard.
        """

    @classmethod
    def from_designation(cls: type[U], designation_substring: str, channels: Channels) -> list[tuple[U, str]]:
        """Parses a designation string into location objects.

        Args:
            designation_substring (str): The full designation string.

            channels (Channels): List of channels to which this color space standard applies.

        Returns:
            list[tuple['Location', Channels, ColorSpaceStandard, str]]: Parsed locations, associated channels, optional
                                                                        color space standards, and remaining substrings.
        """
        if not designation_substring:
            return []

        if designation_substring[0] == '_':
            designation_substring = designation_substring[1:]

        remaining_substring = cls.compatible_with(channels, designation_substring)
        if remaining_substring is not None:
            return [(cls.initialize(channels), remaining_substring)]

        return []

    @staticmethod
    @abstractmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        """Checks whether a designation substring is compatible with the channels.

        Args:
            channels (Channels): The channels to check.
            designation_substring (Optional[str]): The designation substring to validate.

        Returns:
            Optional[str]: Remaining substring if compatible, otherwise None.
        """

    @staticmethod
    def create(designation_substring: str, channels: Channels) -> list[tuple[Optional['ColorSpaceStandard'], str]]:
        """Creates color space standard objects from a designation substring.

        Args:
            designation_substring (str): The substring to parse.
            channels (Channels): The associated channels.

        Returns:
            list[tuple[Optional['ColorSpaceStandard'], str]]: A list of parsed color space standards and remaining
                                                              substrings.
        """
        compos: list[tuple[Optional[ColorSpaceStandard], str]] = [(None, designation_substring)]
        for standard_implementation in all_implementations(ColorSpaceStandard):
            try:
                compos.extend(standard_implementation.from_designation(designation_substring, channels))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {standard_implementation.__name__}')
        return compos


class ItuRBt601(ColorSpaceStandard):
    """Represents the ITU-R BT.601 color space standard."""

    def __init__(self) -> None:
        """Initializes the ITU-R BT.601 color space standard."""
        ColorSpaceStandard.__init__(self, '601')

    @classmethod
    def initialize(cls, channels: Channels) -> 'ItuRBt601':
        """Initializes the Itu.RBt.601 color space standard.

        Args:
            channels (Channels): The channels to associate with this color space standard.

        Returns:
            ItuRBt601: An instance of the Itu.RBt.601 color space standard.
        """
        return cls()

    @staticmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        """Checks if the channels and designation substring are compatible with the Itu.Rbt.601 standard.

        Args:
            channels (Channels): The channels to validate against the standard.
            designation_substring (Optional[str]): The remaining string to check for compatibility.

        Returns:
            Optional[str]: The remaining substring after removing the standard identifier if compatible,
                           otherwise None.
        """
        ident: str = '601'
        if (
            designation_substring is not None and designation_substring.startswith(ident)
        ) and channels.name() == 'YCrCb':
            return designation_substring[len(ident) :]
        return None

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two ItuRBt601 objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, type(self))


class ItuRBt709(ColorSpaceStandard):
    """Represents the ITU-R BT.709 color space standard."""

    def __init__(self) -> None:
        """Initializes the ITU-R BT.709 color space standard."""
        ColorSpaceStandard.__init__(self, '709')

    @classmethod
    def initialize(cls, channels: Channels) -> 'ItuRBt709':
        """Initializes the Itu.RBt.709 color space standard.

        Args:
            channels (Channels): The channels to associate with this color space standard.

        Returns:
            ItuRBt709: An instance of the Itu.RBt.709 color space standard.
        """
        return cls()

    @staticmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        """Checks if the channels and designation substring are compatible with the Itu.RBt.709 standard.

        Args:
            channels (Channels): The channels to validate against the standard.
            designation_substring (Optional[str]): The remaining string to check for compatibility.

        Returns:
            Optional[str]: The remaining substring after removing the standard identifier if compatible,
                           otherwise None.
        """
        ident = '709'
        if (
            designation_substring is not None and designation_substring.startswith(ident)
        ) and channels.name() == 'YCrCb':
            return designation_substring[len(ident) :]

        return None

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two ItuRBt709 objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, type(self))


class Bt601(ColorSpaceStandard):
    """Represents the BT.601 color space standard."""

    def __init__(self) -> None:
        """Initializes the BT.601 color space standard."""
        ColorSpaceStandard.__init__(self, '601')

    @classmethod
    def initialize(cls, channels: Channels) -> 'Bt601':
        """Initializes the BT.601 color space standard.

        Args:
            channels (Channels): The channels to associate with this color space standard.

        Returns:
            Bt601: An instance of the BT.601 color space standard.
        """
        return cls()

    @staticmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        """Checks if the channels and designation substring are compatible with the BT.601 standard.

        Args:
            channels (Channels): The channels to validate against the standard.
            designation_substring (Optional[str]): The remaining string to check for compatibility.

        Returns:
            Optional[str]: The remaining substring after removing the standard identifier if compatible,
                           otherwise None.
        """
        ident = '601'
        if (
            designation_substring is not None and designation_substring.startswith(ident)
        ) and channels.name() == 'YCbCr':
            return designation_substring[len(ident) :]
        return None

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two Bt601 objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, type(self))


class Bt709(ColorSpaceStandard):
    """Represents the BT.709 color space standard."""

    def __init__(self) -> None:
        """Initializes the BT.709 color space standard."""
        ColorSpaceStandard.__init__(self, '709')

    @classmethod
    def initialize(cls, channels: Channels) -> 'Bt709':
        """Initializes the BT.709 color space standard.

        Args:
            channels (Channels): The channels to associate with this color space standard.

        Returns:
            Bt709: An instance of the BT.709 color space standard.
        """
        return cls()

    @staticmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        """Checks if the channels and designation substring are compatible with the BT.709 standard.

        Args:
            channels (Channels): The channels to validate against the standard.
            designation_substring (Optional[str]): The remaining string to check for compatibility.

        Returns:
            Optional[str]: The remaining substring after removing the standard identifier if compatible,
                           otherwise None.
        """
        ident = '709'
        if (
            designation_substring is not None and designation_substring.startswith(ident)
        ) and channels.name() == 'YCbCr':
            return designation_substring[len(ident) :]
        return None

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two Bt601 objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, type(self))


Loc = TypeVar('Loc', bound='Location')


@dataclass
class Location(metaclass=ABCMeta):
    """Abstract base class for pixel locations."""

    @abstractmethod
    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        """Generates a location designation string.

        Args:
            channels (Channels): The associated channels.
            color_space_name (str): The name of the color space.

        Returns:
            str: The location designation.
        """

    @abstractmethod
    def location_name(self, channels: Channels) -> str:
        """Gets the name of the component's location.

        Args:
            channels (Channels): The associated channels.

        Returns:
            str: The location name.
        """

    @abstractmethod
    def number_of_channels(self) -> int:
        """Returns the number of channels in the component.

        This is determined by the length of the component.

        Returns:
            int: The number of channels.
        """

    @staticmethod
    def create(full_designation: str) -> list[tuple[Loc, Channels, Optional[ColorSpaceStandard], str]]:
        """Creates location objects by parsing a full designation string.

        Args:
            full_designation (str): The full string to parse for location designation.

        Returns:
            list[tuple[Loc, Channels, ColorSpaceStandard, str]]: A list of tuples containing:
                - Loc: The created location object.
                - Channels: The associated channels.
                - ColorSpaceStandard: The color space standard associated with the location.
                - str: The remaining designation substring.
        """
        locs: list[tuple[Loc, Channels, Optional[ColorSpaceStandard], str]] = []
        for class_name in all_implementations(Location):
            try:
                locs.extend(class_name.from_designation(full_designation))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {class_name}')
        return locs

    @classmethod
    @abstractmethod
    def from_designation(
        cls: type[Loc], full_designation: str
    ) -> list[tuple[Loc, Channels, Optional[ColorSpaceStandard], str]]:
        """Parses a designation string into location instances.

        Args:
            full_designation (str): The full string to parse for location designation.

        Returns:
            list[tuple[Location, Channels, Optional[ColorSpaceStandard], str]]: A list of tuples containing:
                - Location: The parsed Mono Location object.
                - Channels: The associated channels.
                - Optional[ColorSpaceStandard]: An optional color space standard.
                - str: The remaining designation substring.
        """


@dataclass
class MonoLocation(Location):
    """If a monochrome camera uses one of the mono pixel formats, it outputs 8, 10, 12, or 14 bits of data per pixel.

    If a Basler color camera uses one of the mono pixel formats, the values for each pixel are first converted to the
    YCbCr color model and returns the Y component (brightness).
    This is equivalent to the value that would be derived from a pixel in a monochrome sensor.
    """

    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        """Generates a location designation string.

        Args:
            channels (Channels): The associated channels.
            color_space_name (str): The name of the color space.

        Returns:
            str: The location designation.
        """
        return join_designators(channels.name(), color_space_name)

    def location_name(self, channels: Channels) -> str:
        """Gets the name of the component's location.

        Args:
            channels (Channels): The associated channels.

        Returns:
            str: The location name.
        """
        return 'Mono'

    def number_of_channels(self) -> int:
        """Returns the number of channels in the component.

        This is determined by the length of the component.

        Returns:
            int: The number of channels.
        """
        return 1

    @classmethod
    def from_designation(
        cls: type['MonoLocation'], full_designation: str
    ) -> list[tuple['MonoLocation', Channels, Optional[ColorSpaceStandard], str]]:
        """Parses a designation string into Mono location instances.

        Args:
            full_designation (str): The full string to parse for location designation.

        Returns:
            list[tuple[MonoLocation, Channels, Optional[ColorSpaceStandard], str]]: A list of tuples containing:
                - Location: The parsed Mono Location object.
                - Channels: The associated channels.
                - Optional[ColorSpaceStandard]: An optional color space standard.
                - str: The remaining designation substring.
        """
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]

        potential_channels: list[tuple[Channels, Optional[ColorSpaceStandard], str]] = Channels.create(full_designation)
        mono_locations: list[tuple[MonoLocation, Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, standard, designation_suffix in potential_channels:
            if len(channels) == 1:
                mono_locations.append((MonoLocation(), channels, standard, designation_suffix))
        return mono_locations

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two MonoLocation objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, MonoLocation)


@dataclass
class LMNLocation(Location):
    """Represents an LMN-based pixel location."""

    fields: tuple[int, ...]

    def __post_init__(self) -> None:
        """Performs validation checks on the LMNLocation fields after initialization.

        Raises:
            ValueError: If the fields are empty or have invalid values.
        """
        if len(self.fields) == 0:
            raise ValueError

        if len(self.fields) >= len('LMNOPQRSTUVWXYZ'):
            raise ValueError

        if self.fields[0] != 4:
            raise ValueError

        for field in self.fields[1:]:
            if not (0 < field <= self.fields[0]):
                raise ValueError

    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        """Generates a location designation string.

        Args:
            channels (Channels): The associated channels.
            color_space_name (str): The name of the color space.

        Returns:
            str: The location designation.
        """
        str_channels: str = join_designators(channels.name(), color_space_name)
        if set(self.fields) == {4}:
            return str_channels

        return join_designators(str_channels, ''.join(map(str, self.fields)))

    def location_name(self, channels: Channels) -> str:
        """Gets the name of the component's location.

        Args:
            channels (Channels): The associated channels.

        Returns:
            str: The location name.
        """
        return 'LMNOPQRSTUVWXYZ'[: len(self.fields)] + ''.join(map(str, self.fields))

    def number_of_channels(self) -> int:
        """Returns the number of channels in the component.

        This is determined by the length of the component.

        Returns:
            int: The number of channels.
        """
        return len(self.fields)

    @classmethod
    def from_designation(
        cls: type['LMNLocation'], full_designation: str
    ) -> list[tuple['LMNLocation', Channels, Optional[ColorSpaceStandard], str]]:
        """Parses a designation string into LMN-based location instances.

        Args:
            full_designation (str): The full string to parse for location designation.

        Returns:
            list[tuple[Loc, Channels, Optional[ColorSpaceStandard], str]]: A list of tuples containing:
                - Loc: The parsed LMNLocation object.
                - Channels: The associated channels.
                - Optional[ColorSpaceStandard]: An optional color space standard.
                - str: The remaining designation substring.
        """
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]
        potential_channels: list[tuple[Channels, Optional[ColorSpaceStandard], str]] = Channels.create(full_designation)

        lmn_locations: list[tuple[LMNLocation, Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, standard, designation_suffix in potential_channels:
            number_of_channels = len(channels)
            if len(channels) <= 1:
                continue

            if designation_suffix.startswith('_'):
                designation_suffix = designation_suffix[1:]

            lmn: list[int] = []
            if len(designation_suffix) >= number_of_channels and designation_suffix[0] == '4':
                for i in range(number_of_channels):
                    if designation_suffix[i] in ('1', '2', '4'):
                        lmn.append(int(designation_suffix[i]))
                    else:
                        break

            if len(lmn) == number_of_channels:
                lmn_locations.append(
                    (LMNLocation(tuple(lmn)), channels, standard, designation_suffix[number_of_channels:])
                )
            else:
                lmn_locations.append((LMNLocation((4,) * number_of_channels), channels, standard, designation_suffix))

        return lmn_locations

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two LMNLocation objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, LMNLocation) and self.fields == other.fields


@dataclass
class BayerLocation(Location):
    """Modern color camera sensors are equipped with a Bayer color filter.

    They record red, green, blue pixels in distinct areas and  interpolate in between.
    Interpolation can be proprietary.

    If instead of the interpolated n x m values, you would like to directly receive the bayer
    pattern, use a bayer pixel format. Each pixel is filtered to record only one of the colors.

     array and can output color
    images based on Bayer pixel formats.


    If a color camera uses one of the Bayer pixel formats, it outputs 8, 10, or 12 bits of data per pixel.
    Each pixel is filtered to record only one of the colors red, green, and blue. The pixel data is not processed or
    interpolated in any way. This type of pixel data is sometimes referred to as "raw" output.

    Which Bayer formats are available depends on your camera model.

    If a monochrome camera uses one of the mono pixel formats, it outputs 8, 10, 12, or 14 bits of data per pixel.

    If a Basler color camera uses one of the mono pixel formats, the values for each pixel are first converted to the
    YCbCr color model and returns the Y component (brightness).
    This is equivalent to the value that would be derived from a pixel in a monochrome sensor.
    """

    fields: tuple[str, str, str, str]

    def __post_init__(self) -> None:
        """Validates the BayerLocation fields after initialization.

        Raises:
            ValueError: If the field configuration is invalid.
        """
        if ''.join(sorted(list(set(self.fields)))) != 'LMN':
            raise ValueError

        self.indices: tuple[int, int, int, int] = (
            ord(self.fields[0]) - 76,
            ord(self.fields[1]) - 76,
            ord(self.fields[2]) - 76,
            ord(self.fields[3]) - 76,
        )

    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        """Generates a location designation string.

        Args:
            channels (Channels): The associated channels.
            color_space_name (str): The name of the color space.

        Returns:
            str: The location designation.
        """
        return join_designators(f'Bayer{channels.name(list(self.indices[:2]))}', color_space_name)

    def location_name(self, channels: Channels) -> str:
        """Gets the name of the component's location.

        Args:
            channels (Channels): The associated channels.

        Returns:
            str: The location name.
        """
        return f'Bayer_{"".join(self.fields)}'

    def number_of_channels(self) -> int:
        """Returns the number of channels in the component.

        This is determined by the length of the component.

        Returns:
            int: The number of channels.
        """
        return 3

    @classmethod
    def from_designation(
        cls: type['BayerLocation'], full_designation: str
    ) -> list[tuple['BayerLocation', Channels, Optional[ColorSpaceStandard], str]]:
        """Parses a designation string into BayerLocation instances.

        Args:
            full_designation (str): The string to parse.

        Returns:
            list[tuple[Location, Channels, Optional[ColorSpaceStandard], str]]: Parsed BayerLocation instances,
                associated channels, optional color space standards, and remaining substrings.
        """
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]

        if not full_designation.startswith('Bayer'):
            return []

        full_designation = full_designation[5:]
        potential_channels: list[tuple[Channels, Optional[ColorSpaceStandard], str]] = Channels.create(full_designation)

        bayer_locations: list[tuple[BayerLocation, Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, standard, designation_suffix in potential_channels:
            if len(channels) != 2:
                continue

            # Ground truth; we know that the order is always RGB, and that the Green channel is always the one that
            # is repeated for the bayer location.
            missing_components: list[CompoundChannel] = []
            if CompoundChannel.RED not in channels:
                missing_components.append(CompoundChannel.RED)
            if CompoundChannel.GREEN not in channels:
                missing_components.append(CompoundChannel.GREEN)
            if CompoundChannel.BLUE not in channels:
                missing_components.append(CompoundChannel.BLUE)

            if len(missing_components) != 1 or missing_components[0] is CompoundChannel.GREEN:
                continue

            component_order: dict[Channel, str] = {
                CompoundChannel.RED: 'L',
                CompoundChannel.GREEN: 'M',
                CompoundChannel.BLUE: 'N',
            }
            channel_order: list[str] = [component_order[channels[i]] for i in range(len(channels))]
            missing_component_order: str = component_order[missing_components[0]]

            full_order: list[str] = channel_order + channel_order[::-1]
            if full_order[-1] != 'M':
                full_order[-1] = missing_component_order
            else:
                full_order[-2] = missing_component_order

            assert sorted(set(full_order)) == ['L', 'M', 'N']

            bayer_locations.append(
                (
                    BayerLocation((full_order[0], full_order[1], full_order[2], full_order[3])),
                    CompoundChannels([CompoundChannel.RED, CompoundChannel.GREEN, CompoundChannel.BLUE]),
                    standard,
                    designation_suffix,
                )
            )

        return bayer_locations

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two BayerLocation objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, BayerLocation) and self.fields == other.fields


@dataclass
class BiColor(Location):
    """Represents a Bi-Color pattern pixel location."""

    fields: tuple[str, str, str, str]

    def __post_init__(self) -> None:
        """Processes and validates the BiColor fields during initialization."""
        self.indices: tuple[int, int, int, int] = (
            ord(self.fields[0]) - 76,
            ord(self.fields[1]) - 76,
            ord(self.fields[2]) - 76,
            ord(self.fields[3]) - 76,
        )

    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        """Generates a location designation string.

        Args:
            channels (Channels): The associated channels.
            color_space_name (str): The name of the color space.

        Returns:
            str: The location designation.
        """
        return join_designators(f'BiColor{channels.name(list(self.indices))}', color_space_name)

    def location_name(self, channels: Channels) -> str:
        """Gets the name of the component's location.

        Args:
            channels (Channels): The associated channels.

        Returns:
            str: The location name.
        """
        return f'BiColor_{"".join(self.fields)}'

    def number_of_channels(self) -> int:
        """Returns the number of channels in the component.

        This is determined by the length of the component.

        Returns:
            int: The number of channels.
        """
        return len(set(self.fields))

    @classmethod
    def from_designation(
        cls: type['BiColor'], full_designation: str
    ) -> list[tuple['BiColor', Channels, Optional[ColorSpaceStandard], str]]:
        """Parses a designation string into BiColor instances.

        Args:
            full_designation (str): The string to parse.

        Returns:
            list[tuple[Location, Channels, Optional[ColorSpaceStandard], str]]: Parsed BiColor instances,
                associated channels, optional color space standards, and remaining substrings.
        """
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]
        if not full_designation.startswith('BiColor'):
            return []

        full_designation = full_designation[7:]
        potential_channels: list[tuple[Channels, Optional[ColorSpaceStandard], str]] = Channels.create(
            full_designation, True
        )

        bi_color_locations: list[tuple[BiColor, Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, standard, designation_suffix in potential_channels:
            if len(channels) != 4:
                continue

            component_order: dict[Channel, str] = {
                CompoundChannel.RED: 'L',
                CompoundChannel.GREEN: 'M',
                CompoundChannel.BLUE: 'N',
            }
            try:
                target: list[str] = [component_order[channels[i]] for i in range(len(channels))]
            except KeyError:
                continue

            bi_color_locations.append(
                (
                    BiColor((target[0], target[1], target[2], target[3])),
                    CompoundChannels([CompoundChannel.RED, CompoundChannel.GREEN, CompoundChannel.BLUE]),
                    standard,
                    designation_suffix,
                )
            )

        return bi_color_locations

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two BiColor objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, BiColor) and self.fields == other.fields


@dataclass
class ColorFilterArray(Location, metaclass=ABCMeta):
    """Abstract base class for color filter arrays."""


@dataclass
class SquarePattern(ColorFilterArray, metaclass=ABCMeta):
    """Represents a square pattern color filter array."""

    def __post_init__(self) -> None:
        """Handles post-initialization logic for square pattern color filter arrays.

        Raises:
            NotImplementedError: Indicates the method is not implemented.
        """
        raise NotImplementedError


@dataclass
class SparseColorFilterLocation(ColorFilterArray, metaclass=ABCMeta):
    """A color filter array that includes panchromatic pixels with the red, green and blue color components.

    Different tile patterns can be created.
    """

    fields: tuple[str, str, str, str]

    def __post_init__(self):
        """Validates the sparse color filter fields during initialization.

        Raises:
            ValueError: If the field configuration is invalid.
            NotImplementedError: Indicates the method is not implemented.
        """
        if self.fields[0] != self.fields[2]:
            raise ValueError

        if self.fields[1] == self.fields[3]:
            raise ValueError

        raise NotImplementedError


Comp = TypeVar('Comp', bound='Components')


class Components(metaclass=ABCMeta):
    """Abstract base class for pixel format components."""

    @abstractmethod
    def full_designation(self) -> str:
        """Gets the full designation string for the component.

        Returns:
            str: The full designation string.
        """

    @abstractmethod
    def channel_designation(self) -> str:
        """Gets the designation string for the channels.

        Returns:
            str: The channel designation string.
        """

    @abstractmethod
    def location_name(self) -> str:
        """Gets the name of the component's location.

        Returns:
            str: The location name.
        """

    @property
    def number_of_channels(self) -> int:
        """Returns the number of channels in the component.

        This is determined by the length of the component.

        Returns:
            int: The number of channels.
        """
        return len(self)

    @abstractmethod
    def __len__(self):
        """Gets the number of components.

        Returns:
            int: The number of components.
        """

    @abstractmethod
    def __getitem__(self, item: int) -> Channel:
        """Gets the channel at the specified index.

        Args:
            item (int): The index of the desired channel.

        Returns:
            Channel: The channel at the specified index.
        """

    @staticmethod
    def create(full_designation: str) -> list[tuple['Components', str]]:
        """Parses a full designation string into component objects.

        Iterates through all implementations of the `Components` class, attempting
        to parse the given full designation string into a list of component objects
        and their corresponding substrings.

        Args:
            full_designation (str): The string to parse into components.

        Returns:
            list[tuple[Components, str]]: A list of tuples containing the parsed
            components and their remaining substrings.
        """
        compos: list[tuple[Components, str]] = []
        for implementation in all_implementations(Components):
            try:
                compos.extend(implementation.from_designation(full_designation))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {implementation}')
        return compos

    @classmethod
    @abstractmethod
    def from_designation(cls: type[Comp], full_designation: str) -> list[tuple[Comp, str]]:
        """Parses a designation string into component objects.

        Args:
            full_designation (str): The string to parse.

        Returns:
            list[tuple[Components, str]]: Parsed components and remaining substrings.
        """

    @abstractmethod
    def get_structured_array(self, image_channels: list[np.ndarray]) -> Optional[np.ndarray]:
        """Creates a structured NumPy array from the provided image channels.

        Args:
            image_channels (list[np.ndarray]): A list of image channel arrays. Each array represents one channel.

        Returns:
            Optional[np.ndarray]: A structured NumPy array combining the channels if compatible,
                                  or None if there is a mismatch in shapes or channel numbers.

        Raises:
            ValueError: If channel names or shapes are invalid.
        """


class GenicamComponents(Components):
    """Represents GenICam-compatible pixel format components."""

    def __init__(
        self,
        channels: Union[Channels, Channel, Sequence[Channel]],
        location: Optional[Location] = None,
        color_space_standard: Optional[ColorSpaceStandard] = None,
    ) -> None:
        """Initializes a GenICam-compatible component with channels, location, and color space.

        Args:
            channels (Union[Channels, Channel, Sequence[Channel]]): The channels included in the component.
            location (Optional[Location]): The location information of the component.
            color_space_standard (Optional[ColorSpaceStandard]): The associated color space standard.
        """
        self.channels: Channels
        if is_channel(channels):
            self.channels = CompoundChannels([channels])

        elif isinstance(channels, Sequence):
            self.channels = CompoundChannels(channels)

        else:
            assert isinstance(channels, Channels)
            self.channels = channels

        if not isinstance(self.channels, Channels):
            raise ValueError('Wrong argument')

        self.location: Location
        if location is not None:
            self.location = location
        else:
            if len(self.channels) == 1:
                self.location = MonoLocation()
            else:
                self.location = LMNLocation((4,) * len(self.channels))

        if self.location.number_of_channels() != len(self.channels):
            raise ValueError('Incompatible location')

        self.color_space_standard: Optional[ColorSpaceStandard] = color_space_standard

    def channel_designation(self) -> str:
        """Gets the designation string for the component's channels.

        Returns:
            str: The concatenated name of the channels in the component.
        """
        return self.channels.name()

    def full_designation(self) -> str:
        """Generates the full designation string for the component.

        This includes channel and location information, optionally combined with
        the name of the associated color space standard.

        Returns:
            str: The full designation string of the component.
        """
        color_space_name: str = self.color_space_standard.name if self.color_space_standard is not None else ''
        return self.location.location_designation(self.channels, color_space_name)

    def location_name(self) -> str:
        """Gets the name of the component's location.

        Returns:
            str: The location name.
        """
        return self.location.location_name(self.channels)

    def __len__(self) -> int:
        """Gets the number of channels.

        Returns:
            int: The number of channels.
        """
        return len(self.channels)

    def __getitem__(self, item: int) -> Channel:
        """Gets the channel at the specified index.

        Args:
            item (int): The index of the desired channel.

        Returns:
            Channel: The channel at the specified index.
        """
        return self.channels[item]

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two GenicamComponent objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        if isinstance(other, GenicamComponents):
            return (
                self.channels == other.channels
                and self.location == other.location
                and self.color_space_standard == other.color_space_standard
            )
        return False

    @classmethod
    def from_designation(
        cls: type['GenicamComponents'], full_designation: str
    ) -> list[tuple['GenicamComponents', str]]:
        """Parses a designation string into GenicamComponents instances.

        Args:
            full_designation (str): The full string representing the designation.

        Returns:
            list[tuple[Components, str]]: A list of tuples containing GenicamComponents and the remaining substring.

        Logs:
            Warnings are logged if exceptions occur during parsing by location implementations.
        """
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]
        compos: list[tuple[GenicamComponents, str]] = []
        for location_implementation in all_implementations(Location):
            try:
                channels_location = location_implementation.from_designation(full_designation)
                for location, channels, standard, substring in channels_location:
                    compos.append((GenicamComponents(channels, location, standard), substring))

            except Exception as e:
                logger.warning(f'An exception occurred: {e} {location_implementation.__name__}')
        return compos

    def get_structured_array(self, image_channels: list[np.ndarray]) -> Optional[np.ndarray]:
        """Creates a structured NumPy array from the provided image channels.

        Args:
            image_channels (list[np.ndarray]): A list of image channel arrays. Each array represents one channel.

        Returns:
            Optional[np.ndarray]: A structured NumPy array combining the channels if compatible,
                                  or None if there is a mismatch in shapes or channel numbers.

        Raises:
            ValueError: If channel names or shapes are invalid.
        """
        dtypes: list[tuple] = []

        if len(image_channels) != len(self.channels):
            return None

        for channel, image_channel in zip(self.channels, image_channels):
            name: str = channel_name(channel)
            dtypes.append((name, image_channel.dtype))

            if image_channel.shape != image_channels[0].shape:
                return None

        image_structured = np.empty(image_channels[0].shape, dtype=dtypes)
        for (key, _type), image in zip(dtypes, image_channels):
            image_structured[key] = image

        return image_structured


class CustomComponents(Components, metaclass=ABCMeta):
    """Abstract base class for custom pixel format components."""


def bayer_rgb(
    location: tuple[str, str, str, str], color_space_standard: Optional[ColorSpaceStandard] = None
) -> GenicamComponents:
    """Creates a Bayer RGB component.

    Args:
        location (tuple[str, str, str, str]): The Bayer pattern location.
        color_space_standard (Optional[ColorSpaceStandard]): The associated color space standard.

    Returns:
        GenicamComponents: The Bayer RGB component.
    """
    return GenicamComponents(
        [CompoundChannel.RED, CompoundChannel.GREEN, CompoundChannel.BLUE],
        BayerLocation(location),
        color_space_standard=color_space_standard,
    )


def bi_color_rgb(
    location: tuple[str, str, str, str], color_space_standard: Optional[ColorSpaceStandard] = None
) -> GenicamComponents:
    """Creates a Bi-Color RGB component.

    Args:
        location (tuple[str, str, str, str]): The Bi-Color pattern location.
        color_space_standard (Optional[ColorSpaceStandard]): The associated color space standard.

    Returns:
        GenicamComponents: The Bi-Color RGB component.
    """
    return GenicamComponents(
        [CompoundChannel.RED, CompoundChannel.GREEN, CompoundChannel.BLUE],
        BiColor(location),
        color_space_standard=color_space_standard,
    )


class SizedColorSpaceStandard(metaclass=ABCMeta):
    """Abstract base class for size-dependent color space standards."""

    def __init__(self, name: str) -> None:
        """Initializes a sized color space standard with a name."""
        self.name: str = name

    @classmethod
    @abstractmethod
    def from_designation(
        cls, designation_substring: str, comps: Components, sizes: Size
    ) -> list[tuple['SizedColorSpaceStandard', str]]:
        """Parses a designation string into a sized color space standard.

        Args:
            designation_substring (str): The string to parse.
            comps (Components): The associated components.
            sizes (Size): The associated size information.

        Returns:
            list[tuple[SizedColorSpaceStandard, str]]: Parsed standards and remaining substrings.
        """

    @staticmethod
    @abstractmethod
    def compatible_with(comps: Components, sizes: Size) -> bool:
        """Checks if the components and size are compatible with the standard.

        Args:
            comps (Components): The components to check.
            sizes (Size): The size information to validate.

        Returns:
            bool: True if compatible, False otherwise.
        """

    @staticmethod
    def create(
        designation_substring: str, comps: Components, sizes: Size
    ) -> list[tuple[Optional['SizedColorSpaceStandard'], str]]:
        """Creates sized color space standard objects from a designation substring.

        Args:
            designation_substring (str): The substring to parse.
            comps (Components): The associated components.
            sizes (Size): The associated size information.

        Returns:
            list[tuple[Optional[SizedColorSpaceStandard], str]]: Parsed standards and remaining substrings.
        """
        compos: list[tuple[Optional[SizedColorSpaceStandard], str]] = [(None, designation_substring)]
        for standard_implementation in all_implementations(SizedColorSpaceStandard):
            try:
                compos.extend(standard_implementation.from_designation(designation_substring, comps, sizes))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {standard_implementation.__name__}')
        return compos


class RGB10V1(SizedColorSpaceStandard):
    """Represents the RGB10 V1 color space standard."""

    def __init__(self) -> None:
        """Initializes the RGB10 V1 color space standard."""
        SizedColorSpaceStandard.__init__(self, 'V1')

    @staticmethod
    def compatible_with(comps: Components, sizes: Size) -> bool:
        """Checks if the components and size are compatible with the RGB10 V1 standard.

        Args:
            comps (Components): The components to check.
            sizes (Size): The size information to validate.

        Returns:
            bool: True if compatible, False otherwise.
        """
        return comps.channel_designation() == 'RGB' and sizes.designation() == '10'

    @classmethod
    def from_designation(
        cls, designation_substring: str, comps: Components, sizes: Size
    ) -> list[tuple[SizedColorSpaceStandard, str]]:
        """Parses a designation string into an RGB10 V1 color space standard.

        Args:
            designation_substring (str): The string to parse.
            comps (Components): The associated components.
            sizes (Size): The associated size information.

        Returns:
            list[tuple[SizedColorSpaceStandard, str]]: Parsed standards and remaining substrings.
        """
        if not designation_substring:
            return []

        if not cls.compatible_with(comps, sizes):
            return []

        if designation_substring.startswith('V1'):
            return [(cls(), designation_substring[len('V1') :])]
        return []

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two SizedColorSpaceStandard objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, RGB10V1)


class RGB12V1(SizedColorSpaceStandard):
    """Represents the RGB12 V1 color space standard."""

    def __init__(self) -> None:
        """Initializes the RGB12 V1 color space standard."""
        SizedColorSpaceStandard.__init__(self, 'V1')

    @staticmethod
    def compatible_with(comps: Components, sizes: Size) -> bool:
        """Checks if the components and size are compatible with the RGB12 V1 standard.

        Args:
            comps (Components): The components to check.
            sizes (Size): The size information to validate.

        Returns:
            bool: True if compatible, False otherwise.
        """
        return comps.channel_designation() == 'RGB' and sizes.designation() == '12'

    @classmethod
    def from_designation(
        cls, designation_substring: str, comps: Components, sizes: Size
    ) -> list[tuple[SizedColorSpaceStandard, str]]:
        """Parses a designation string into an RGB12 V1 color space standard.

        Args:
            designation_substring (str): The string to parse.
            comps (Components): The associated components.
            sizes (Size): The associated size information.

        Returns:
            list[tuple[SizedColorSpaceStandard, str]]: Parsed standards and remaining substrings.
        """
        if not designation_substring:
            return []

        if not cls.compatible_with(comps, sizes):
            return []

        if designation_substring.startswith('V1'):
            return [(cls(), designation_substring[len('V1') :])]
        return []

    def __eq__(self, other: Any) -> bool:
        """Checks equality between two RGB12V1 objects.

        Args:
            other (Any): The object to compare.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        return isinstance(other, RGB12V1)
