"""Module for handling the size of pixel formats.

This module provides the `Size` class for representing the size of pixel formats
and methods for converting between designations and size representations.
"""

from typing import Optional
from dataclasses import dataclass


@dataclass
class Size:
    """Represents the size of a pixel format.

    Attributes:
        sizes (list[int]): A list of sizes for each channel.
    """

    sizes: list[int]

    def __len__(self) -> int:
        """Returns the number of sizes.

        Returns:
            int: The number of sizes.
        """
        return len(self.sizes)

    def __getitem__(self, item) -> int:
        """Gets the size at the specified index.

        Args:
            item (int): The index of the size to retrieve.

        Returns:
            int: The size at the specified index.
        """
        return self.sizes[item]

    def __eq__(self, other) -> bool:
        """Checks if two Size instances are equal.

        Args:
            other: The other Size instance to compare.

        Returns:
            bool: True if the instances are equal, False otherwise.
        """
        return isinstance(other, Size) and self.sizes == other.sizes

    def designation(self) -> str:
        """Computes the designation string for the Size instance.

        Returns:
            str: The designation string.
        """
        if not len(self.sizes):
            return ''

        num_sizes: set[int] = set(self.sizes)
        if len(num_sizes) <= 1:
            return str(self.sizes[0])

        max_size: int = max(self.sizes)
        max_size_space: int = len(str(max_size))
        return ''.join([str(size).zfill(max_size_space) for size in self.sizes])

    @staticmethod
    def from_designation(designation: str, number_of_channels: int) -> Optional[tuple['Size', str]]:
        """Creates a Size instance from a designation string.

        While this can be ambiguous, wer can assume some prevalent sizes.

        Args:
            designation (str): The designation string.
            number_of_channels (int): The number of channels.

        Returns:
            Optional[tuple[Size, str]]: A tuple containing the Size instance and the remaining substring.
        """
        if number_of_channels == 0:
            return Size([]), designation

        if not len(designation):
            return None

        if designation[0] == '_':
            designation = designation[1:]

        number_substring: str = ''
        for i in designation:
            if not i.isdigit():
                break
            number_substring += i
        remaining_substring = designation[len(number_substring) :]

        len_number_substring: int = len(number_substring)
        if not len_number_substring:
            return None

        # In this case, it is obviously the same size for every channel.
        if len_number_substring % number_of_channels != 0:
            return Size(
                [
                    int(number_substring),
                ]
                * number_of_channels
            ), remaining_substring

        # It is unclear if it is the same size for every channel,
        number_of_digits_per_channel = len_number_substring // number_of_channels

        channels: list[int] = []
        for i_channel in range(number_of_channels):
            size = int(
                number_substring[
                    number_of_digits_per_channel * i_channel : number_of_digits_per_channel * (i_channel + 1)
                ]
            )
            channels.append(size)

        all_suitable = len(str(max(channels))) == number_of_digits_per_channel

        # It is either: channels or number_of_channels * (int(number_substring), ).
        if number_substring.startswith('0') or number_of_channels == 1 or all_suitable:
            return Size(channels), remaining_substring

        return Size(
            [
                int(number_substring),
            ]
            * number_of_channels
        ), remaining_substring
