"""This module provides structures representing sensor states, configurations, and sensor lifecycle management.

It includes classes for defining sensors, their states, configurations, and event handling mechanisms.
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional, Union, TYPE_CHECKING, Callable

from dataspree.inference.sensor.data.exceptions import DriverFeatureImplementationError, DriverInvalidState

from dataspree.inference.sensor.data.configuration import NotAvailable, Auto, OptionalConfiguration, NotAvailableWrite

if TYPE_CHECKING:
    from dataspree.inference.sensor.data.sensor_callback import SensorCallback


class SensorState(Enum):
    """Contains all the states in a sensor livecycle."""

    UNAVAILABLE = 0
    DISCOVERED = 1
    CONNECTED = 2
    STARTED = 3


class SensorStateTransition(Enum):
    """Contains all the state transitions."""

    STOPPED = -2
    DISCONNECTED = -1
    LOST = 0
    FOUND = 1
    CONNECTED = 2
    STARTED = 3


@dataclass
class TemperatureControl(OptionalConfiguration):
    """Configuration for temperature control of a sensor.

    Attributes:
        target_temperature (Optional[Union[int, NotAvailable]]): The target temperature setting.
        fan (Optional[Union[Auto, int, NotAvailable]]): The fan setting, which can be automatic or manual.
        active_cooling (Optional[Union[Auto, int, NotAvailable]]): The active cooling setting, which can be automatic
                                                                   or manual.
    """

    target_temperature: Optional[Union[int, NotAvailable]] = field(default_factory=NotAvailable)
    fan: Optional[Union[Auto, int, NotAvailable]] = field(default_factory=NotAvailable)
    active_cooling: Optional[Union[Auto, int, NotAvailable]] = field(default_factory=NotAvailable)


@dataclass
class Sensor(OptionalConfiguration):
    """Sensor representation.

    Args:
        identifier (Union[str, NotAvailableWrite]):
            Identifier that must be unique. Is used to identify a particular device. Can also be used to determine
            the sensor model if a model for a particular device is registered.
            If this is None, the sensor instance is just used for supplying configurations to one or more actual
            sensors.

        internal_device_id (Union[int, NotAvailable])
            identifier for the driver (if necessary). Might change with new cameras connecting, so we cannot use this
            value reliably outside the driver.

        model (Union[str, NotAvailable]):
            Sensor model. Is determined the best fitting sensor mode.

        manufacturer (Union[str, NotAvailable]):
            Sensor manufacturer. Example: Basler. Is used to determine the best fitting model.

        name (Union[str, NotAvailable]):
            User defined name.

        firmware_version (Union[str, NotAvailable]):
            Firmware version gathered from the device.

        temperature_control (Union[NotAvailable, TemperatureControl]):
            Temperature control settings.

        state (SensorState):
            Current state of the sensor.

        device_properties (Dict[str, Any]):
            Additional properties that are relevant to the sensor.
    """

    identifier: Union[str, NotAvailableWrite]
    internal_device_id: Union[int, NotAvailable] = field(default_factory=NotAvailable)
    model: Union[str, NotAvailable] = field(default_factory=NotAvailable)
    manufacturer: Union[str, NotAvailable] = field(default_factory=NotAvailable)
    name: Union[str, NotAvailable] = field(default_factory=NotAvailable)
    firmware_version: Union[str, NotAvailable] = field(default_factory=NotAvailable)

    temperature_control: Union[NotAvailable, TemperatureControl] = field(default_factory=NotAvailable)

    state: SensorState = SensorState.UNAVAILABLE
    device_properties: dict[str, Any] = field(default_factory=dict)
    __callback: Optional['SensorCallback'] = None
    __initialized: bool = field(init=False)

    def __post_init__(self) -> None:
        """Initializes the sensor to ensure it's fully set up after instantiation."""
        self.__initialized = True

    def register_callback(self, cb: 'SensorCallback') -> None:
        """Registers a callback for the sensor.

        Args:
            cb (SensorCallback): The callback instance to register.

        Raises:
            DriverInvalidState: If a callback is already registered.
        """
        if self.__callback is not None:
            raise DriverInvalidState

        self.__callback = cb

    def __report_sensor_state_transition(self, target_state: SensorState, backward: bool) -> None:
        """Internal function that is used to dispatch to the required sensor state transition callback.

        Raises:
        DriverFeatureImplementationError: if the callback raises.
        """
        try:
            if self.__callback is None:
                raise AttributeError
            lookup_dict: dict[Optional[SensorState], Callable[[Sensor], None]] = (
                {
                    SensorState.UNAVAILABLE: self.__callback.on_sensor_lost,
                    SensorState.DISCOVERED: self.__callback.on_sensor_disconnected,
                    SensorState.CONNECTED: self.__callback.on_sensor_stopped,
                }
                if backward
                else {
                    SensorState.DISCOVERED: self.__callback.on_sensor_found,
                    SensorState.CONNECTED: self.__callback.on_sensor_connected,
                    SensorState.STARTED: self.__callback.on_sensor_started,
                }
            )
        except AttributeError:
            raise DriverFeatureImplementationError from None
        else:
            try:
                lookup_dict[target_state](self)
            except KeyError as e:
                raise DriverFeatureImplementationError(f'Unable to invoke callback: {e}') from None

            except DriverFeatureImplementationError as e:
                raise e from None

            except Exception as e:
                raise DriverFeatureImplementationError(
                    f'Invalid exception raised by callback: {type(e).__name__} {e}'
                ) from e

        finally:
            super().__setattr__('sensor', target_state)

    def __setattr__(self, key: str, value: Any) -> None:
        """Sets an attribute value, enforcing restrictions on specific attributes.

        Args:
            key (str): The attribute name.
            value (Any): The value to set.

        Raises:
            AttributeError: If attempting to change the identifier after initialization.

            DriverFeatureImplementationError
        """
        if key == 'identifier':
            try:
                if self.__initialized and self.identifier is not None:
                    raise ValueError('Cannot change identifier.')

            except ValueError as e:
                raise AttributeError(str(e))

            except AttributeError:
                # if __initialized or self.identifier don't exist yet.
                pass

        if key == 'state':
            try:
                initialized = self.__initialized
            except AttributeError:
                initialized = False

            if initialized:
                for event in range(self.state.value + 1, value.value + 1):
                    try:
                        self.__report_sensor_state_transition(SensorState(event), False)
                    except DriverFeatureImplementationError as e:
                        raise e from None

                for event in list(range(value.value, self.state.value))[::-1]:
                    try:
                        self.__report_sensor_state_transition(SensorState(event), True)
                    except DriverFeatureImplementationError as e:
                        raise e from None

        OptionalConfiguration.__setattr__(self, key, value)

    def __str__(self) -> str:
        """Returns the string representation of the sensor.

        Returns:
            str: The identifier of the sensor.
        """
        return str(self.identifier)


@dataclass
class SensorData:
    """Basic sensor data representation.

    Arguments:
         timestamp_inference (int): the timestamp of creation / acquisition.
    """

    timestamp_inference: int
