import copy
import time
import traceback

import numpy as np

from abc import abstractmethod
from typing import TypeVar, Optional, Union, Callable
from logging import getLogger

from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.core.item import Item
from dataspree.inference.sensor.data.camera import (
    Camera,
    CameraData,
    AcquisitionMode,
    TriggerType,
    TriggerMode,
    SoftwareTrigger,
    HardwareTrigger,
    TriggerActivation,
    TriggerAcquisitionMode,
    ContinuousAcquisitionMode,
    ManualGain,
    Roi,
    AutoGain,
    Brightness,
    TimedExposure,
    TriggeredExposure,
    AutoExposure,
    WhiteBalance,
    ManualWhiteBalance,
    AutoWhiteBalance,
    CheckerPatternRemoval,
    TonalRange,
    Binning,
    ContrastAdjustments,
    ColorAdjustment,
    TemperatureState,
    Gamma,
    NotAvailable,
    AcquisitionStatus,
    ExposureStatus,
    FrameBurstStatus,
)
from dataspree.inference.sensor.data.exceptions import (
    NoFreeHardwareTrigger,
    NoFreeSoftwareTrigger,
    ConfigurationInconsistent,
    SensorException,
    SensorFeaturePermanentlyNotAvailable,
    DriverFeatureNotImplemented,
    SensorFeatureTemporarilyNotAvailable,
    InvalidState,
    InputException,
    SensorInvalidState,
    ConfigurationError,
    ResourceError,
    SensorFeatureUnavailable,
    DriverFeatureImplementationError,
)
from dataspree.inference.sensor.data.pixel_format import PixelFormat, GenicamPixelFormat
from dataspree.inference.sensor.data.configuration import NotSupported, NotAvailableRead, is_available
from dataspree.inference.sensor.driver.sensor_driver import SensorDriver

T = TypeVar('T', bound=Camera)

logger = getLogger(__name__)


class CameraDriver(SensorDriver[T]):
    """A base class for a camera driver with essential methods and callback handling."""

    def refresh_attributes(self, sensor: T) -> None:
        """Refresh all available attributes.

        Arguments:
            sensor (T): the sensor for which the attributes should be refreshed

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        SensorDriver.refresh_attributes(self, sensor)

        refresh_functions: list[Callable[[Camera], None]] = [
            self.refresh_acquisition_mode,
            self.refresh_operation_mode,
            self.refresh_pixel_format,
            self.refresh_current_camera_configuration_identifier,
            self.refresh_startup_camera_configuration_identifier,
            self.refresh_resulting_effective_acquisition_rate,
            self.refresh_crop,
            self.refresh_binning,
            self.refresh_brightness,
            self.refresh_white_balance,
            self.refresh_contrast_adjustments,
            self.refresh_color_adjustment,
            self.refresh_gamma,
            self.refresh_blooming_reduction,
            self.refresh_pattern_removal,
            self.refresh_tonal_range,
            self.refresh_rois,
            self.refresh_camera_model,
        ]
        sensor = self._sensor(sensor)
        for fun in refresh_functions:
            try:
                fun(sensor)
            except SensorInvalidState as e:
                logger.debug(f'{fun.__name__} cannot be performed in the current sensor state {e}.')

            except (InvalidState, ConfigurationError) as e:
                logger.warning(f'{fun.__name__} raised an {type(e).__name__} error: {e}.')
                if is_application_debug():
                    traceback.print_exc()

            except ResourceError as e:
                logger.debug(f'{fun.__name__} raised a resource error {e}.')

            except SensorFeatureTemporarilyNotAvailable as e:
                logger.debug(f'{fun.__name__} temporarily not available {e}.')

            except (SensorFeatureUnavailable, DriverFeatureNotImplemented):
                pass

            except DriverFeatureImplementationError as e:
                logger.warning(f'{fun.__name__} implementation error{e}.')
                if is_application_debug():
                    traceback.print_exc()

            except InputException as e:
                logger.error(f'{fun.__name__} failed: {e}.')
                if is_application_debug():
                    traceback.print_exc()

    def set_attributes(self, sensor: T, bounds_check: bool = False) -> None:
        """Apply all attributes that were provided.

        Fails when the first exception is thrown.

        Arguments:
            sensor (T):
                A sensor configuration that should be applied, that has all the fields.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        SensorDriver.set_attributes(self, sensor, bounds_check=bounds_check)

        old_configuration = self.sensors[sensor.identifier]

        setter = [
            (old_configuration.acquisition_mode, sensor.acquisition_mode, self.set_acquisition_mode),
            (old_configuration.operation_mode, sensor.operation_mode, self.set_operation_mode),
            (old_configuration.pixel_format, sensor.pixel_format, self.set_pixel_format),
            (
                old_configuration.current_camera_configuration_identifier,
                sensor.current_camera_configuration_identifier,
                self.set_current_camera_configuration_identifier,
            ),
            (
                old_configuration.startup_camera_configuration_identifier,
                sensor.startup_camera_configuration_identifier,
                self.set_startup_camera_configuration_identifier,
            ),
            (old_configuration.crop, sensor.crop, self.set_crop),
            (old_configuration.binning, sensor.binning, self.set_binning),
            (old_configuration.brightness, sensor.brightness, self.set_brightness),
            (old_configuration.white_balance, sensor.white_balance, self.set_white_balance),
            (old_configuration.contrast_adjustments, sensor.contrast_adjustments, self.set_contrast_adjustments),
            (old_configuration.color_adjustment, sensor.color_adjustment, self.set_color_adjustment),
            (old_configuration.gamma, sensor.gamma, self.set_gamma),
            (old_configuration.blooming_reduction, sensor.blooming_reduction, self.set_blooming_reduction),
            (old_configuration.pattern_removal, sensor.pattern_removal, self.set_checker_pattern_removal),
            (old_configuration.tonal_range, sensor.tonal_range, self.set_tonal_range),
            (old_configuration.rois, sensor.rois, self.set_rois),
        ]

        for old_config_value, new_config_value, func in setter:
            try:
                func(sensor, new_config_value, bounds_check=bounds_check)
            except InputException as e:
                if not old_config_value.entails(new_config_value):
                    raise e

    def refresh_camera_model(self, sensor: T) -> None:
        """Refresh the camera model.

        Arguments:
            sensor (T): the sensor for which the attributes should be refreshed

        Raises:
            SensorNotFound if the sensor is currently not found.
        """

    #
    # Acquisition rate
    #

    def fetch_resulting_effective_acquisition_rate(self, sensor: T) -> Union[float, NotAvailableRead]:  # noqa
        """Return a camera's acquisition rate.

        Arguments:
            sensor (T): The sensor from which we gather the acquisition rate.

        Returns:
            Union[float, NotAvailableRead]: the camera's acquisition rate.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def refresh_resulting_effective_acquisition_rate(self, sensor: T) -> None:
        """Fetch and update the effective acquisition rate.

        Arguments:
            sensor (T): The sensor from which we gather and update the frame rate.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.resulting_effective_acquisition_rate = self.fetch_resulting_effective_acquisition_rate(sensor)

    #
    # Acquisition Mode
    #

    @abstractmethod
    def fetch_acquisition_mode(self, sensor: T) -> Union[AcquisitionMode, NotAvailableRead]:
        """Retrieve the currently active acquisition mode from the specified sensor.

        Args:
            sensor (T): The sensor object from which to retrieve the acquisition mode.

        Returns:
            Union[AcquisitionMode, NotAvailableRead]: The current acquisition mode of the sensor.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """

    def refresh_acquisition_mode(self, sensor: T) -> None:
        """Fetch and update the currently active acquisition mode from the specified sensor.

        Args:
            sensor (T): The sensor object from which to retrieve the acquisition mode.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.acquisition_mode = self.fetch_acquisition_mode(sensor)

    @abstractmethod
    def set_acquisition_mode(
        self, sensor: T, value: Union[AcquisitionMode, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Set the acquisition mode for the specified sensor.

        Args:
            sensor (T): The sensor object for which the acquisition mode is to be set.
            value (Union[AcquisitionMode, NotAvailable]): The acquisition mode to be applied to the sensor.
            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """

    def configure_acquisition_max_rate(
        self, sensor: T, value: Optional[Union[float, NotAvailable]], bounds_check: bool = False
    ) -> None:
        """Set the maximum acquisition rate for the specified sensor.

        Utility function.

        Args:
            sensor (T):
                The camera object for which the maximum acquisition rate is to be set.

            value (Optional[Union[float, NotAvailable]]):
                The maximum acquisition rate to be applied. If `None`, rate limitation is disabled.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of
                1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        acquisition_mode = copy.deepcopy(sensor.acquisition_mode)
        acquisition_mode.max_rate = value
        self.set_acquisition_mode(sensor, acquisition_mode, bounds_check=bounds_check)

    def configure_acquisition_continuous(
        self, sensor: T, max_acquisition_rate: Optional[Union[float, NotAvailable]] = None, bounds_check: bool = False
    ) -> None:
        """Configure the specified sensor for continuous acquisition mode.

        Utility function.

        Sets the acquisition mode of the given sensor to continuous acquisition. This mode captures frames continuously
        without acquisition triggers. Exposure triggers are not affected by this, and might still be required.

        Args:
            sensor (T):
                The camera object to be configured for continuous acquisition.

            max_acquisition_rate (Optional[Union[float, NotAvailable]]):
                The maximum acquisition rate to be applied. If `None`, rate limitation is disabled.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of
                1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)

        acquisition_mode = ContinuousAcquisitionMode(max_acquisition_rate)
        if not sensor.acquisition_mode.entails(acquisition_mode):
            logger.debug(f'Switching from {sensor.acquisition_mode} to {acquisition_mode}.')
            self.set_acquisition_mode(sensor, acquisition_mode, bounds_check=bounds_check)

    def configure_acquisition_trigger_based(
        self,
        sensor: T,
        frame_start: TriggerType = TriggerType.SOFTWARE,
        frame_stop: Optional[Union[TriggerType, NotAvailable]] = None,
        max_acquisition_rate: Optional[Union[int, NotAvailable]] = None,
        bounds_check: bool = False,
    ) -> None:
        """Configure the specified sensor for trigger-based acquisition mode.

        Utility function.

        Sets the acquisition mode of the given sensor to trigger-based acquisition. Configures the triggers for starting
        and stopping frame capture based on the provided trigger types.

        Args:
            sensor (T):
                The camera object to be configured for trigger-based acquisition.

            frame_start (TriggerType, default=TriggerType.SOFTWARE):
                The trigger type for starting frame capture.

            frame_stop (Optional[TriggerType], default=None):
                The trigger type for stopping frame capture.

            max_acquisition_rate (Optional[int]):
                The maximum acquisition rate to be applied. If `None`, rate limitation is disabled.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of
                1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        # Build Acquisition Mode:
        # 1. Get free triggers
        free_software_triggers = sensor.get_free_software_triggers()
        free_hardware_triggers = sensor.get_free_hardware_triggers()
        if isinstance(sensor.acquisition_mode, TriggerAcquisitionMode):
            for trigger in (sensor.acquisition_mode.trigger_frame_end, sensor.acquisition_mode.trigger_start):
                if sensor.is_software_trigger(trigger):
                    free_software_triggers.insert(0, trigger.trigger_source)
                elif sensor.is_hardware_trigger(sensor.acquisition_mode.trigger_frame_end):
                    free_hardware_triggers.insert(0, trigger.trigger_source)

        # 2. Set frame start trigger
        trigger_frame_start: TriggerMode
        if frame_start is TriggerType.SOFTWARE:
            try:
                trigger_name = sensor.camera_model.software_trigger_frame_start or free_software_triggers.pop()
                trigger_frame_start = SoftwareTrigger(trigger_name, 0)
            except IndexError:
                raise NoFreeSoftwareTrigger.create()
        else:
            try:
                trigger_frame_start = HardwareTrigger(free_hardware_triggers.pop(), 0, TriggerActivation.RISING_EDGE)
            except IndexError:
                raise NoFreeHardwareTrigger.create()

        # 3. Set frame stop trigger
        trigger_frame_stop: Optional[TriggerMode] = None
        if frame_stop is TriggerType.SOFTWARE:
            try:
                trigger_frame_stop = SoftwareTrigger(free_software_triggers.pop(), 0)
            except IndexError:
                raise NoFreeSoftwareTrigger.create()

        elif frame_stop is TriggerType.HARDWARE:
            trigger_frame_stop = HardwareTrigger(trigger_frame_start.trigger_source, 0, TriggerActivation.FALLING_EDGE)

        # 4. Either patch the trigger acquisition mode or initialize it, depending on the currently active mode.
        acquisition_mode: AcquisitionMode
        if not isinstance(sensor.acquisition_mode, TriggerAcquisitionMode):
            acquisition_mode = TriggerAcquisitionMode(
                max_acquisition_rate, trigger_start=trigger_frame_start, trigger_frame_end=trigger_frame_stop
            )
        else:
            acquisition_mode = copy.copy(sensor.acquisition_mode)
            acquisition_mode.max_rate = max_acquisition_rate
            acquisition_mode.trigger_frame_start = trigger_frame_start
            acquisition_mode.trigger_frame_end = trigger_frame_stop

        # Attempt to apply the acquisition mode.
        if not sensor.acquisition_mode.entails(acquisition_mode):
            logger.debug(f'Switching from {sensor.acquisition_mode} to {acquisition_mode}.')
            self.set_acquisition_mode(sensor, acquisition_mode, bounds_check=bounds_check)

    def configure_acquisition_frame_burst(
        self,
        sensor: T,
        start: Optional[TriggerType] = TriggerType.SOFTWARE,
        stop: Optional[Union[TriggerType, NotAvailable]] = None,
        burst_frame_count: Optional[Union[int, NotAvailable]] = None,
        burst_mode: Optional[Union[str, NotAvailable]] = None,
        bounds_check: bool = False,
    ) -> None:
        """Configure the sensor for frame burst acquisition mode.

        Utility function.

        This method sets up the sensor to capture a series of frames in a burst mode, where each burst is triggered
        by specified start and stop triggers. It also allows for specifying the number of frames in the burst and the
        mode of burst acquisition.
        This method can only be used if the sensor is configured to run in trigger-based acquisition mode.

        Args:
            sensor (T):
                The camera object to be configured for frame burst acquisition.

            start (Optional[TriggerType], default=TriggerType.SOFTWARE):
                The trigger type to start the frame burst.
                Can be either `TriggerType.SOFTWARE` or `TriggerType.HARDWARE`.

            stop (Optional[TriggerType], default=None):
                The trigger type to stop the frame burst.
                Can be either `TriggerType.SOFTWARE` or `TriggerType.HARDWARE`.
                If `None`, the burst will not have a defined stop trigger.

            burst_frame_count (Optional[int], default=None):
                The number of frames to capture in each burst. If `None`, the frame count is not modified.

            burst_mode (Optional[str], default=None):
                The mode of burst acquisition, which may be specific to the camera model. If `None`,
                the burst mode is not modified.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of
                1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)

        if not isinstance(sensor.acquisition_mode, TriggerAcquisitionMode):
            raise ConfigurationInconsistent(
                'Frame burst acquisition can only be ' 'activated after activating trigger acquisition mode!'
            )

        # Build Acquisition Mode:
        # 1. Get free triggers
        free_software_triggers = sensor.get_free_software_triggers()
        free_hardware_triggers = sensor.get_free_hardware_triggers()
        for trigger in (
            sensor.acquisition_mode.trigger_frame_burst_end,
            sensor.acquisition_mode.trigger_frame_burst_start,
        ):
            if sensor.is_software_trigger(trigger):
                free_software_triggers.insert(0, trigger.trigger_source)
            elif sensor.is_hardware_trigger(sensor.acquisition_mode.trigger_frame_end):
                free_hardware_triggers.insert(0, trigger.trigger_source)

        # 2. Set start trigger
        trigger_frame_burst_start: Optional[TriggerMode] = None
        if start is TriggerType.SOFTWARE:
            try:
                trigger_frame_burst_start = SoftwareTrigger(free_software_triggers.pop(), 0)
            except IndexError:
                raise NoFreeSoftwareTrigger.create()
        elif start is TriggerType.HARDWARE:
            try:
                trigger_frame_burst_start = HardwareTrigger(
                    free_hardware_triggers.pop(), 0, TriggerActivation.RISING_EDGE
                )
            except IndexError:
                raise NoFreeHardwareTrigger.create()

        # 3. Set stop trigger
        trigger_frame_burst_stop: Optional[TriggerMode] = None
        if stop is TriggerType.SOFTWARE:
            try:
                trigger_frame_burst_stop = SoftwareTrigger(free_software_triggers.pop(), 0)
            except IndexError:
                raise NoFreeSoftwareTrigger.create()
        elif stop is TriggerType.HARDWARE:
            trigger_frame_burst_stop = HardwareTrigger(
                trigger_frame_burst_start.trigger_source, 0, TriggerActivation.FALLING_EDGE
            )

        acquisition_mode = copy.copy(sensor.acquisition_mode)
        acquisition_mode.trigger_frame_burst_start = trigger_frame_burst_start
        acquisition_mode.trigger_frame_end = trigger_frame_burst_stop
        if burst_frame_count is not None:
            acquisition_mode.burst_frame_count = burst_frame_count
        if burst_mode is not None:
            acquisition_mode.burst_mode = burst_mode

        # Attempt to apply the acquisition mode.
        self.set_acquisition_mode(sensor, acquisition_mode, bounds_check=bounds_check)

    #
    # Acquisition status (not stored in sensor)
    #

    def fetch_acquisition_status(self, sensor: T) -> Union[AcquisitionStatus, NotAvailableRead]:  # noqa
        """Returns the acquisition status if possible.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Union[AcquisitionStatus, NotAvailableRead]: the acquisition status

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def fetch_exposure_status(self, sensor: T) -> Union[ExposureStatus, NotAvailableRead]:  # noqa
        """Returns the exposure status if possible.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Union[ExposureStatus, NotAvailableRead]: the exposure status

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def fetch_frame_burst_status(self, sensor: T) -> Union[FrameBurstStatus, NotAvailableRead]:  # noqa
        """Returns the frame burst status if possible.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Union[FrameBurstStatus, NotAvailableRead]: the frame burst status

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    #
    # Triggering
    #

    def execute_trigger(self, sensor: T, timestamp: Optional[int] = None) -> None:
        """Execute a trigger for acquiring data.

        This defaults to the frame start trigger.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            timestamp (Optional[int]): Timestamp of the trigger

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        self.execute_software_trigger_frame(sensor, True, timestamp=timestamp)

    @abstractmethod
    def execute_software_trigger_frame(self, sensor: T, start: bool, timestamp: Optional[int] = None) -> None:
        """Execute a software start or stop trigger for triggering a frame.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            start (bool): frame start / frame stop.

            timestamp (Optional[int]): Timestamp of the trigger

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """

    def execute_software_trigger_frame_burst(self, sensor: T, start: bool) -> None:
        """Execute a software start or stop trigger for triggering a frame.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            start (bool): frame start / frame stop.

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        raise DriverFeatureNotImplemented

    def execute_software_trigger_exposure(self, sensor: T, start: bool) -> None:
        """Execute te software start or stop trigger that controls exposure.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            start (bool): frame exposure / stop exposure.

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        raise DriverFeatureNotImplemented

    def execute_software_trigger(
        self, sensor: T, trigger: Optional[Union[TriggerMode, str]], timestamp: Optional[int] = None
    ) -> None:
        """Execute te software start or stop trigger that controls exposure.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            trigger (Optional[Union[TriggerMode, str]]): Trigger that is supposed to be executed.

            timestamp (Optional[int]): Trigger that is supposed to be executed.

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        raise DriverFeatureNotImplemented

    def configure_output_line_value(self, sensor: T, output_line_name: str, trigger_value: bool) -> None:
        """Set output line to high / low.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            output_line_name (str): Name of the output line.

            trigger_value (bool): value the trigger should assume.

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        raise DriverFeatureNotImplemented

    #
    # Regions Of Interest
    #

    def highlight_roi(self, sensor: T, roi: str, highlight: bool) -> None:
        """Visually highlights a region of interest for configuration purpose.

        Arguments:
            sensor (T): The camera from which we gather the information.

            roi (str): the name of the region of interest

            highlight (bool) whether to highlight the ROI in the image.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    def fetch_roi(self, sensor: T, roi: str) -> Union[Roi, NotAvailable]:
        """Retrieve roi from camera.

        Arguments:
            sensor (T): The camera from which we gather the information.

            roi (str): the name of the region of interest

        Returns:
            Union[Roi, NotAvailable] the region of interest if available.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailable()

    def fetch_rois(self, sensor: T) -> dict[str, Union[Roi, NotAvailable]]:
        """Retrieve roi from camera.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Dict[str, Union[Roi, NotAvailable]] the regions of interest

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        return {roi: self.fetch_roi(sensor, roi) for roi in sensor.camera_model.rois}

    def refresh_rois(self, sensor: T) -> None:
        """Fetch and update the camera's rois.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor.rois = self.fetch_rois(self._sensor(sensor))

    def configure_roi(
        self, sensor: T, name: str, roi: Optional[Union[Roi, NotAvailable]], bounds_check: bool = False
    ) -> None:
        """Configure an auto function ROI.

        The settings for the auto function rois are independent of the settings for the image roi.
        There should be no effect of the general ROI onto the configured auto rois.

        Attributes:
            sensor (T): The sensor, to which the roi settings are to be applied.

            roi (str): The name of the region of interest that is to be configured.

            roi: Optional[Union[Roi, NotAvailable]]:
                The region of interest that is to be configured.

                If this value is None, defaults to the entire image.

                Attention: Don't use None if you configured a camera crop,
                and want to use the same pixels as in the camera crop.
                Use `sensor.crop` as argument instead!

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    def set_rois(
        self, sensor: T, value: dict[str, Union[Roi, Optional[NotAvailable]]], bounds_check: bool = False
    ) -> None:
        """Configure auto function ROIs.

        Arguments:
            sensor (T): The camera from which we gather the information.

            value (Dict[str, Union[Roi, Optional[NotAvailable]]): the auto function rois.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        for roi_name, roi_value in value.items():
            self.configure_roi(sensor, roi_name, roi_value, bounds_check=bounds_check)
        sensor.rois = value

    #
    # Brightness
    #

    def fetch_brightness(self, sensor: T) -> Union[Brightness, NotAvailable]:
        """Return a camera's brightness settings.

        Returns:
            Union[Brightness, NotAvailable] the camera's brightness settings if they could be retrieved.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailable()

    def refresh_brightness(self, sensor: T) -> None:
        """Fetch and update the camera's brightness settings.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.brightness = self.fetch_brightness(sensor)
        if sensor.brightness.auto_function_enabled and sensor.brightness.auto_roi_configuration is None:
            logger.warning(
                'Invalid brightness configuration: Auto functionality is enabled, but no region '
                'of interest is provided.'
            )

    def set_brightness(self, sensor: T, value: Union[Brightness, NotAvailable], bounds_check: bool = False):
        """Configure brightness.

        Arguments:
            sensor (T): The camera to which we apply the settings.

            value (Union[Brightness, NotAvailable]): The brightness configuration.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Gain
    #

    def configure_gain_manual(
        self, sensor: T, value: Union[float, int] = 0, digital: Optional[bool] = None, bounds_check: bool = False
    ) -> None:
        """Configure manual gain.

        Utility function.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            value: Union[float, int]: The configured gain. On some camera models, the gain must be entered as a "raw"
                                      value on an integer scale. The camera needs the raw value for its internal
                                      processing mechanism. The raw value, however, isn't the same as the actual gain
                                      value, which is expressed in decibels (dB).

            digital: Optional[bool]: Distinguishes between analog and digital gain.
                                     Analog gain is applied before the signal from the camera sensor is converted into
                                     digital values. Digital gain is applied after the conversion, i.e., it is basically
                                     a multiplication of the digitized values.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.gain = ManualGain(digital=digital, value=value)
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def configure_gain_auto(
        self,
        sensor: T,
        upper_limit: Optional[Union[float, int]] = None,
        lower_limit: Optional[Union[float, int]] = None,
        auto_target_brightness: Optional[float] = None,
        roi_configuration: Optional[Union[str, Roi]] = None,
        digital: Optional[bool] = None,
        bounds_check: bool = False,
    ) -> None:
        """Configure automatic gain.

        Utility function.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            digital: Optional[bool]: Distinguishes between analog and digital gain.
                                     Analog gain is applied before the signal from the camera sensor is converted into
                                     digital values. Digital gain is applied after the conversion, i.e., it is basically
                                     a multiplication of the digitized values.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

            upper_limit: Union[float, int]: Upper limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.
            lower_limit: Union[float, int]: Lower limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.

            auto_target_brightness: Optional[float]: Brightness that is to be attained for automatic gain control.

            roi_configuration: Optional[Union[str, Roi]] Roi which is used to assess gain configuration.
                                                         If empty and the sensor supports roi configuration, use the
                                                         entire image as roi

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.gain = AutoGain(digital=digital, upper_limit=upper_limit, lower_limit=lower_limit, temporary=False)
        if auto_target_brightness:
            brightness.auto_target_brightness = auto_target_brightness
        if roi_configuration:
            brightness.auto_roi_configuration = roi_configuration
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def configure_gain_manual_through_automatic_finder(
        self,
        sensor: T,
        upper_limit: Optional[Union[float, int]] = None,
        lower_limit: Optional[Union[float, int]] = None,
        auto_target_brightness: Optional[float] = None,
        roi_configuration: Optional[Union[str, Roi]] = None,
        digital: Optional[bool] = None,
        bounds_check: bool = False,
    ) -> None:
        """Run automatic gain once, and use the determined values for manual gain mode.

        Utility function.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            digital: Optional[bool]: Distinguishes between analog and digital gain.
                                     Analog gain is applied before the signal from the camera sensor is converted into
                                     digital values. Digital gain is applied after the conversion, i.e., it is basically
                                     a multiplication of the digitized values.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

            upper_limit: Union[float, int]: Upper limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.
            lower_limit: Union[float, int]: Lower limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.

            auto_target_brightness: Optional[float]: Brightness that is to be attained for automatic gain control.

            roi_configuration: Optional[Union[str, Roi]] Roi which is used to assess gain configuration.
                                                         If empty and the sensor supports roi configuration, use the
                                                         entire image as roi

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.gain = AutoGain(digital=digital, upper_limit=upper_limit, lower_limit=lower_limit, temporary=True)
        brightness.auto_target_brightness = auto_target_brightness
        brightness.auto_roi_configuration = roi_configuration
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

        time_start = time.time()
        brightness: Brightness
        while True:
            if (time.time() - time_start) >= 5.0:
                raise SensorFeatureTemporarilyNotAvailable('Timeout')
            brightness = self.fetch_brightness(sensor)
            if isinstance(brightness.gain, AutoGain):
                if not brightness.gain.temporary:
                    raise SensorException('Unexpected state retrieved.')
                time.sleep(0.5)
            elif brightness.gain is None:
                raise SensorFeaturePermanentlyNotAvailable
            else:
                break

        sensor.brightness = brightness

    #
    # Exposure
    #

    def configure_exposure_timed(self, sensor: T, exposure_time: float, bounds_check: bool = False) -> None:
        """Configure timed exposure.

        Arguments:
            sensor (T): The camera to which we apply the settings.

            exposure_time (float) exposure time in microseconds.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.exposure = TimedExposure(exposure_time)
        if not sensor.brightness.exposure.entails(brightness.exposure):
            self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def configure_exposure_triggered(
        self,
        sensor: T,
        trigger_exposure_start: TriggerMode,
        trigger_exposure_end: Optional[TriggerMode] = None,
        start_delay: Optional[int] = None,
        overlap_time_max: Optional[float] = None,
        bounds_check: bool = False,
    ) -> None:
        """Configure triggered exposure.

        Arguments:
            sensor (T): The camera to which we apply the settings.

            trigger_exposure_start: Optional[TriggerMode] trigger that either indicates exposure start, or exposure
                                                          start and stop if the trigger has a
                                                          duration.

            trigger_exposure_end: Optional[TriggerMode] trigger that indicates stop of the exposure.

            start_delay: Optional[int]:  Determines the period of time between the detection of the trigger signal and
                                         the actual start of the exposure.

            overlap_time_max: Optional[float]: With overlapping image acquisition, exposure of the next image begins
                                               when the camera is reading out the sensor data of the previous image.
                                               In triggered exposure mode, the camera does not know how long the image
                                               is exposed. Overlap_time_max can be used to indicate the shortest
                                               exposure time that is intended to use be used in microseconds.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)

        brightness = copy.deepcopy(sensor.brightness)
        brightness.exposure = TriggeredExposure(
            trigger_exposure_start=trigger_exposure_start,
            trigger_exposure_end=trigger_exposure_end,
            start_delay=start_delay,
            overlap_time_max=overlap_time_max,
        )
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def configure_exposure_auto(
        self,
        sensor: T,
        upper_limit: Optional[Union[float, int]] = None,
        lower_limit: Optional[Union[float, int]] = None,
        auto_target_brightness: Optional[float] = None,
        roi_configuration: Optional[Union[str, Roi]] = None,
        bounds_check: bool = False,
    ) -> None:
        """Configure automatic gain.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

            upper_limit: Union[float, int]: Upper limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.
            lower_limit: Union[float, int]: Lower limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.

            auto_target_brightness: Optional[float]: Brightness that is to be attained for automatic gain control.

            roi_configuration: Optional[Union[str, Roi]] Roi which is used to assess gain configuration.
                                                         If empty and the sensor supports roi configuration, use the
                                                         entire image as roi

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.exposure = AutoExposure(upper_limit=upper_limit, lower_limit=lower_limit, temporary=False)
        if auto_target_brightness:
            brightness.auto_target_brightness = auto_target_brightness
        if roi_configuration:
            brightness.auto_roi_configuration = roi_configuration
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def configure_exposure_manual_through_automatic_finder(
        self,
        sensor: T,
        upper_limit: Optional[Union[float, int]] = None,
        lower_limit: Optional[Union[float, int]] = None,
        auto_target_brightness: Optional[float] = None,
        roi_configuration: Optional[Union[str, Roi]] = None,
        bounds_check: bool = False,
    ) -> None:
        """Run automatic exposure once, and use the determined values for timed exposure mode.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

            upper_limit: Union[float, int]: Upper limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.
            lower_limit: Union[float, int]: Lower limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.

            auto_target_brightness: Optional[float]: Brightness that is to be attained for automatic gain control.

            roi_configuration: Optional[Union[str, Roi]] Roi which is used to assess gain configuration.
                                                         If empty and the sensor supports roi configuration, use the
                                                         entire image as roi

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.exposure = AutoExposure(upper_limit=upper_limit, lower_limit=lower_limit, temporary=True)
        if auto_target_brightness:
            brightness.auto_target_brightness = auto_target_brightness
        if roi_configuration:
            brightness.auto_roi_configuration = roi_configuration
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

        brightness: Brightness
        time_start = time.time()
        while True:
            if (time.time() - time_start) >= 5.0:
                raise SensorFeatureTemporarilyNotAvailable('Timeout')
            brightness = self.fetch_brightness(sensor)
            if isinstance(brightness.exposure, AutoExposure):
                if not brightness.exposure.temporary:
                    raise SensorException('Unexpected state retrieved.')
                time.sleep(0.5)
            elif isinstance(brightness.exposure, TimedExposure):
                break
            else:
                raise SensorFeaturePermanentlyNotAvailable

        sensor.brightness = brightness

        self.refresh_effective_acquisition_rate(sensor)

    #
    # White balance
    #

    def fetch_white_balance(self, sensor: T) -> Union[WhiteBalance, NotAvailable]:
        """Return the white balance settings.

        Arguments:
            sensor (T): The sensor from which we gather the white balance.

        Returns:
            Optional[float]: the camera's white balance.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailable()

    def refresh_white_balance(self, sensor: T) -> None:
        """Retrieve and apply the white balance settings.

        Arguments:
            sensor (T): The sensor from which we gather the white balance.

        Returns:
            Optional[float]: the camera's white balance.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.white_balance = self.fetch_white_balance(sensor)

    def set_white_balance(
        self, sensor: T, value: Union[WhiteBalance, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Apply the white balance settings.

        Arguments:
            sensor (T): The sensor to which apply the white balance settings

            value (Union[WhiteBalance, NotAvailable]): The white balance settings

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    def configure_white_balance_manual(
        self, sensor: T, balance_ratios: Optional[dict[str, float]] = None, bounds_check: bool = False
    ) -> None:
        """Apply manual white balance.

        Arguments:
            sensor (T): The sensor to which apply the white balance settings

            balance_ratios (Dict[str, float]): Contains the balance ratio for each channel. Typically,
                the balance ratios are configured as follows:
                balance_ratio == 1: The intensity of the color remains unchanged.
                               > 1: The intensity of the color is increased relative to the other two colors.
                               < 1: The intensity of the color is decreased relative to the other two colors.
                However, some cameras don't allow balance ratios smaller than one, because a balance
                can be normalized to an equivalent balance by dividing through the sum of all
                balance ratios or multiplying each value with a specific factor.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        # TODO: we could think about replacing 1.0 by the sum (?) of all values.
        #       This also needs to be tested with a RGB camera.
        sensor = self._sensor(sensor)
        full_ratios = {channel: 1.0 for channel in sensor.camera_model.white_balance_manual_channels}
        full_ratios.update(balance_ratios or dict())
        self.set_white_balance(sensor, ManualWhiteBalance(full_ratios), bounds_check=bounds_check)

    def configure_white_balance_auto(
        self,
        sensor: T,
        roi_configuration: Optional[Union[str, Roi]] = None,
        adjustment_damping: Optional[float] = None,
        bounds_check: bool = False,
    ) -> None:
        """Configure automatic white balance finder.

        Arguments:
            sensor (T): The sensor to which apply the white balance settings

            roi_configuration (Optional[Union[str, Roi]]): The roi on which the auto white balance is executed

            adjustment_damping (Optional[float]):  If specified, colors are not adjusted immediately, but after a
                certain delay. This can be useful to avoid the automatic control mechanism of the auto functions being
                disrupted by objects moving in and out of the camera's area of view.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)
        self.set_white_balance(
            sensor, AutoWhiteBalance(roi_configuration, adjustment_damping, temporary=False), bounds_check=bounds_check
        )

    def configure_white_balance_manual_through_automatic_finder(
        self,
        sensor: T,
        roi_configuration: Optional[Union[str, Roi]] = None,
        adjustment_damping: Optional[float] = None,
        bounds_check: bool = False,
    ) -> None:
        """Run automatic white balance finder once, and thus determine and use parameters for manual white balance mode.

        Arguments:
            sensor (T): The sensor to which apply the white balance settings

            roi_configuration (Optional[Union[str, Roi]]): The roi on which the auto white balance is executed

            adjustment_damping (Optional[float]):  If specified, colors are not adjusted immediately, but after a
                certain delay. This can be useful to avoid the automatic control mechanism of the auto functions being
                disrupted by objects moving in and out of the camera's area of view.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        sensor = self._sensor(sensor)

        self.set_white_balance(
            sensor, AutoWhiteBalance(roi_configuration, adjustment_damping, temporary=True), bounds_check=bounds_check
        )

        time_start = time.time()
        white_balance: WhiteBalance
        while True:
            if (time.time() - time_start) >= 5.0:
                raise SensorFeatureTemporarilyNotAvailable('Timeout')
            white_balance = self.fetch_white_balance(sensor)
            if isinstance(white_balance, AutoWhiteBalance):
                if not white_balance.temporary:
                    raise SensorException('Unexpected auto white balance state retrieved.')
                time.sleep(0.5)
            elif isinstance(white_balance, ManualWhiteBalance):
                break
            else:
                raise SensorFeatureTemporarilyNotAvailable

        sensor.white_balance = white_balance

    #
    # Checker pattern removal
    #

    def fetch_pattern_removal(self, sensor: T) -> Optional[Union[CheckerPatternRemoval, NotAvailable]]:  # noqa
        """Return the checker pattern removal settings.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[Union[CheckerPatternRemoval]]: the camera's checker pattern removal configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailable()

    def refresh_pattern_removal(self, sensor: T) -> None:
        """Retrieve and apply the checker pattern removal settings.

        Arguments:
            sensor (T): The sensor from which we gather the checker pattern removal.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.pattern_removal = self.fetch_pattern_removal(sensor)

    def set_pattern_removal(
        self, sensor: T, value: Optional[Union[CheckerPatternRemoval, NotAvailable]], bounds_check: bool = False
    ) -> None:
        """Apply the checker pattern removal settings.

        Arguments:
            sensor (T):
                The sensor to which apply the checker pattern removal settings.

            value (Optional[Union[CheckerPatternRemoval, NotAvailable]]):
                The checker pattern removal settings.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of
                1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Tonal Range Configuration
    #

    def fetch_tonal_range(self, sensor: T) -> Optional[Union[TonalRange, NotAvailableRead]]:  # noqa
        """Return the tonal range configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[Union[TonalRange, NotAvailableRead]]: the camera's tonal range configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def refresh_tonal_range(self, sensor: T) -> None:
        """Retrieve and apply the tonal range configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.tonal_range = self.fetch_tonal_range(sensor)

    def set_tonal_range(
        self, sensor: T, value: Optional[Union[TonalRange, NotAvailable]], bounds_check: bool = False
    ) -> None:
        """Apply the tonal range configuration.

        Arguments:
            sensor (T): The sensor to which apply the checker pattern removal settings.

            value (Optional[Union[TonalRange, NotAvailable]]): The checker pattern removal settings.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Binning
    #

    def fetch_binning(self, sensor: T) -> Union[Binning, NotAvailableRead]:
        """Return the binning configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Union[Binning, NotAvailableRead]: the camera's binning configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def refresh_binning(self, sensor: T) -> None:
        """Retrieve and apply the binning configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.binning = self.fetch_binning(sensor)

    def set_binning(self, sensor: T, value: Union[Binning, NotAvailable], bounds_check: bool = False) -> None:
        """Apply the binning configuration.

        Arguments:
            sensor (T): The sensor to which apply the binning configuration.

            value (Union[Binning, NotAvailable]): The binning configuration.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Blooming reduction
    #
    def fetch_blooming_reduction(self, sensor: T) -> Union[bool, NotAvailableRead]:
        """Return the blooming reduction configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Union[bool, NotAvailableRead]: the T's blooming reduction configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def refresh_blooming_reduction(self, sensor: T) -> None:
        """Retrieve and apply the blooming reduction configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.blooming_reduction = self.fetch_blooming_reduction(sensor)

    def set_blooming_reduction(self, sensor: T, value: Union[bool, NotAvailable], bounds_check: bool = False) -> None:
        """Apply the blooming reduction configuration.

        Arguments:
            sensor (T): The sensor to which apply the blooming reduction configuration.

            value (Union[bool, NotAvailable]): The blooming reduction configuration.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    def set_checker_pattern_removal(
        self, sensor: T, value: Union[bool, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Apply the checker pattern removal.

        Arguments:
            sensor (T): The sensor to which apply the blooming reduction configuration.

            value (Union[bool, NotAvailable]): The blooming reduction configuration.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Contrast Adjustments
    #

    def fetch_contrast_adjustments(self, sensor: T) -> Optional[Union[ContrastAdjustments, NotAvailableRead]]:
        """Return the contrast adjustment configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Union[ContrastAdjustments, NotAvailable]: the T's contrast adjustment configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def refresh_contrast_adjustments(self, sensor: T) -> None:
        """Retrieve and apply the contrast adjustment configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.contrast_adjustments = self.fetch_contrast_adjustments(sensor)

    def set_contrast_adjustments(
        self, sensor: T, value: Optional[Union[ContrastAdjustments, NotAvailable]], bounds_check: bool = False
    ) -> None:
        """Apply the contrast adjustment configuration.

        Arguments:
            sensor (T):
                The sensor to which apply the configuration.

            value (Optional[Union[ContrastAdjustments, NotAvailable]]):
                The contrast adjustment configuration.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of
                1000.0 even though the maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Operation mode
    #

    def fetch_operation_mode(self, sensor: T) -> Union[str, NotAvailableRead]:
        """Return the operation mode.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Union[str, NotAvailableRead]: The sensor's operation mode.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def refresh_operation_mode(self, sensor: T) -> None:
        """Retrieve and apply the operation mode.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.operation_mode = self.fetch_operation_mode(sensor)

    def set_operation_mode(
        self,
        sensor: T,
        value: Union[str, NotAvailable],
        bounds_check: bool = False,
    ) -> None:
        """Apply the operation mode.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            value (Union[str, NotAvailable]): the new operation mode

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Color adjustment
    #

    def fetch_color_adjustment(self, sensor: T) -> Optional[Union[ColorAdjustment, NotAvailableRead]]:
        """Return the color adjustment.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[Union[ColorAdjustment, NotAvailable]]: The sensor's color adjustment.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def refresh_color_adjustment(self, sensor: T) -> None:
        """Retrieve and apply the color adjustment.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.color_adjustment = self.fetch_color_adjustment(sensor)

    def set_color_adjustment(
        self, sensor: T, value: Optional[Union[ColorAdjustment, NotAvailableRead]], bounds_check: bool = False
    ) -> None:
        """Apply the color adjustment.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            value (Optional[Union[ColorAdjustment, NotAvailable]]): the new color adjustment.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Gamma
    #

    def fetch_gamma(self, sensor: T) -> Optional[Union[Gamma, NotAvailableRead]]:  # noqa
        """Return the gamma settings.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[Union[Gamma, NotAvailable]]: The sensor's gamma configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailableRead()

    def refresh_gamma(self, sensor: T) -> None:
        """Retrieve and apply the gamma configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.gamma = self.fetch_gamma(sensor)

    def set_gamma(self, sensor: T, value: Gamma, bounds_check: bool = False) -> None:
        """Apply the gamma configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            value (Optional[Union[Gamma, NotAvailableRead]]): the new gamma configuration.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Device temperature
    #

    def fetch_device_temperature(self, sensor: T) -> Union[float, NotAvailableRead]:
        """Return the current device temperature.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Union[float, NotAvailable]: The sensor's temperature.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotSupported()

    def fetch_device_temperature_state(self, sensor: T) -> Union[TemperatureState, NotAvailableRead]:
        """Return the current device temperature state.

        The Temperature State camera indicates whether the camera's internal temperature is normal or too high.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[float]: The sensor's temperature state

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotSupported()

    #
    # Pixel format
    #

    @abstractmethod
    def fetch_pixel_format(self, sensor: T) -> Union[PixelFormat, NotAvailableRead]:
        """Return the current device pixel format.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Union[PixelFormat, NotAvailableRead]: The sensor's pixel format.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """

    def refresh_pixel_format(self, sensor: T) -> None:
        """Retrieve and apply the pixel format.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.pixel_format = self.fetch_pixel_format(sensor)

    @abstractmethod
    def set_pixel_format(self, sensor: T, value: Union[PixelFormat, NotAvailable], bounds_check: bool = False) -> None:
        """Apply the pixel format.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            value (PixelFormat): the new pixel format.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """

    #
    # Crop
    #

    def fetch_crop(self, sensor: T) -> Optional[Union[Roi, NotAvailable]]:
        """Return the current crop.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[Union[Roi, NotAvailable]]: The sensor's crop.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotSupported()

    def refresh_crop(self, sensor: T) -> None:
        """Retrieve and apply the crop.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.crop = self.fetch_crop(sensor)

    def set_crop(self, sensor: T, value: Optional[Union[Roi, NotAvailable]] = None, bounds_check: bool = False) -> None:
        """Apply the crop.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            value (Optional[Union[Roi, NotAvailable]]):
                The new crop. None to set to largest possible crop.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Camera configurations
    #

    def fetch_current_camera_configuration_identifier(self, sensor: T) -> Union[int, str, NotAvailableRead]:
        """Return the current camera configuration.

        This does not convey any information on whether the current settings are actually loaded from the camera
        configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Union[int, str, NotAvailableRead]: The sensor's camera configuration identifier.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotSupported()

    def refresh_current_camera_configuration_identifier(self, sensor: T) -> None:
        """Retrieve and apply the current camera configuration name.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.current_camera_configuration_identifier = self.fetch_current_camera_configuration_identifier(sensor)

    def set_current_camera_configuration_identifier(
        self,
        sensor: T,
        value: Union[int, str, NotAvailable],
        bounds_check: bool = False,
    ) -> None:
        """Sent the current camera configuration to the camera configuration that is identified
        by the provided name. This does not automatically load the configuration, as we might
        want to set the current camera configuration just to apply our current configuration to it.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            value (Union[int, str, NotAvailable]): the camera configuration to apply / load.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    def load_current_camera_configuration(self, sensor: T) -> None:
        """Load the current configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        raise DriverFeatureNotImplemented

    def save_current_camera_configuration(self, sensor: T) -> None:
        """Save the current configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        raise DriverFeatureNotImplemented

    def fetch_startup_camera_configuration_identifier(self, sensor: T) -> Union[str, int, NotAvailableRead]:  # noqa
        """Return the startup camera configuration name.

        This does not convey any information on whether the current settings are actually loaded from the camera
        configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Union[str, int, NotAvailableRead]: The sensor's camera configuration identifier.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotSupported()

    def refresh_startup_camera_configuration_identifier(self, sensor: T) -> None:
        """Retrieve and apply the camera startup configuration name.

        This does not convey any information on whether the current settings are actually loaded from the camera
        configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        sensor.startup_camera_configuration_identifier = self.fetch_startup_camera_configuration_identifier(sensor)

    def set_startup_camera_configuration_identifier(
        self, sensor: T, value: Union[str, int, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Sent the startup camera configuration to the camera configuration that is identified
        by the provided name. This does not automatically load the configuration, as we might
        want to set the current camera configuration just to apply our current configuration to it.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            value (Union[str, int, NotAvailable]): the camera configuration to apply / load.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    def load_startup_camera_configuration(self, sensor: T) -> None:
        """Load the configured startup configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        self.set_current_camera_configuration_identifier(sensor, sensor.startup_camera_configuration_identifier)
        self.load_current_camera_configuration(sensor)

    def save_current_configuration_as_startup_configuration(
        self, sensor: T, value: Optional[Union[str, int, NotAvailable]] = None
    ) -> None:
        """Save the current configuration as startup configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            value (Optional[str]): optionally provide a new startup configuration name.
                                                               If None, use the currently configured startup config.

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        if is_available(value):
            self.set_startup_camera_configuration_identifier(sensor, value)

        self.set_current_camera_configuration_identifier(sensor, sensor.startup_camera_configuration_identifier)
        self.save_current_camera_configuration(sensor)

    #
    # Data acquisition
    #

    @abstractmethod
    def fetch_sensor_data(self, sensor: T, timeout: float = 0.0) -> Union[CameraData, NotAvailable]:
        """Fetches data from the specified sensor, waiting up to a defined timeout.

        Args:
            sensor (T): The sensor from which to retrieve data.

            timeout (float, optional): The maximum time in seconds to wait for data. Defaults to 0.0.

        Returns:
            Union[SensorData, NotAvailable]: The sensor data if available; otherwise, a NotAvailable status.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            ResourceError: If the requested resource is not available.

            FeatureUnavailable: If the feature is not available for the sensor or the driver.
        """

    def create_item(self, sensor: T, sensor_data: CameraData) -> Item:
        """Creates an item based on sensor data and associates it with the specified sensor.

        Args:
            sensor (T): The sensor for which the item is created.
            sensor_data (SensorData): The data used to create the item.

        Returns:
            Item: An item that contains all the relevant sensory information from the sensor data.

        Raises:
            ResourceError: If the requested resource is not available.

            DriverFeatureImplementationError: implementation error
        """
        sensor = self._sensor(sensor)
        pixel_format = sensor.pixel_format

        try:
            image_structured = pixel_format.get_structured_array(sensor_data.content)
        except Exception as e:
            logger.warning(e)
            image_structured = None

        image: Optional[np.ndarray] = None
        if isinstance(pixel_format, GenicamPixelFormat) and len(sensor_data.content) > 1:
            channel_designation = pixel_format.component.channel_designation()

            # Specific presets.
            channel_designations: dict[str, tuple[int, ...]] = {
                'RGB': (0, 1, 2),
                'YRGB': (0, 1, 2, 3),
                'RGBa': (0, 1, 2, 3),
                'aRGB': (0, 1, 2, 3),
                'HSI': (0, 1, 2),
                'HSV': (0, 1, 2),
                'YCbCr': (0, 1, 2),
                'aBGR': (0, 1, 2, 3),
                'BGRa': (0, 1, 2, 3),
            }

            designation = channel_designations.get(channel_designation, tuple(range(len(pixel_format.component))))

            fallback = not len(designation)
            for idx in designation:
                if 0 <= idx < len(sensor_data.content):
                    first = sensor_data.content[designation[0]]
                    current = sensor_data.content[idx]
                    fallback = not (current.shape == first.shape) or not (current.dtype == first.dtype)
                else:
                    fallback = True

                if fallback:
                    break

            if not fallback:
                first = sensor_data.content[designation[0]]

                image = np.empty((*first.shape, len(designation)), first.dtype)
                for i, idx in enumerate(designation):
                    image[..., i] = sensor_data.content[idx]

                if image.shape[-2] == 1:
                    image = image[..., 0, :]

        if image is None:
            image = sensor_data.content[0] if len(sensor_data.content) else None

        try:
            sensor_copy = copy.deepcopy(sensor)
        except Exception:
            sensor_copy = None

        trigger_timestamp = sensor.trigger_timestamp
        sensor.trigger_timestamp = None

        return Item.default(
            id=f'{sensor.identifier}_',
            image=image,
            image_structured=image_structured,
            image_channels=sensor_data.content,
            camera_identifier=sensor.identifier,
            camera_model=sensor.camera_model,
            camera_manufacturer=sensor.manufacturer,
            camera_name=sensor.name,
            sensor=sensor_copy,
            gain=sensor_data.gain,
            exposure_time=sensor_data.exposure_time,
            sensor_data=sensor_data,
            trigger_count=sensor.trigger_count,
            timestamp_trigger=trigger_timestamp,
            timestamp_acquisition=sensor.acquisition_timestamp,
        )
